import os

import angr


test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


def test_patch_vulnerable_fauxware_amd64():
    binpath = os.path.join(test_location, "x86_64", "vulns", "vulnerable_fauxware")
    proj = angr.Project(binpath, auto_load_libs=False)

    proj.kb.patches.add_patch(0x40094C, b"\x0a")
    patched = proj.kb.patches.apply_patches_to_binary()

    # manual patch
    with open(binpath, "rb") as f:
        binary_data = f.read()
    binary_data = binary_data[:0x94C] + b"\x0a" + binary_data[0x94D:]

    assert patched == binary_data


if __name__ == "__main__":
    test_patch_vulnerable_fauxware_amd64()



import os


import angr
from angr.knowledge_plugins.xrefs import XRef, XRefType

test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


def test_lwip_udpecho_bm():
    bin_path = os.path.join(test_location, "armel", "lwip_udpecho_bm.elf")
    p = angr.Project(bin_path, auto_load_libs=False)
    cfg = p.analyses.CFG(data_references=True)

    func = cfg.functions[0x23C9]
    state = p.factory.blank_state()
    prop = p.analyses.Propagator(func=func, base_state=state)
    _ = p.analyses.XRefs(func=func, replacements=prop.replacements)

    timenow_cp_xrefs = p.kb.xrefs.get_xrefs_by_dst(0x23D4)  # the constant in the constant pool
    timenow_xrefs = p.kb.xrefs.get_xrefs_by_dst(0x1FFF36F4)  # the value in .bss

    assert len(timenow_cp_xrefs) == 1
    assert next(iter(timenow_cp_xrefs)) == XRef(ins_addr=0x23C9, dst=0x23D4, xref_type=XRefType.Read)

    assert len(timenow_xrefs) == 3
    assert [x for x in timenow_xrefs if x.type == XRefType.Offset][0] == XRef(
        ins_addr=0x23C9, dst=0x1FFF36F4, xref_type=XRefType.Offset
    )
    assert [x for x in timenow_xrefs if x.type == XRefType.Read][0] == XRef(
        ins_addr=0x23CB, dst=0x1FFF36F4, xref_type=XRefType.Read
    )
    assert [x for x in timenow_xrefs if x.type == XRefType.Write][0] == XRef(
        ins_addr=0x23CF, dst=0x1FFF36F4, xref_type=XRefType.Write
    )


def test_lwip_udpecho_bm_the_better_way():
    bin_path = os.path.join(test_location, "armel", "lwip_udpecho_bm.elf")
    p = angr.Project(bin_path, auto_load_libs=False)
    p.analyses.CFG(cross_references=True)

    timenow_cp_xrefs = p.kb.xrefs.get_xrefs_by_dst(0x23D4)  # the constant in the constant pool
    timenow_xrefs = p.kb.xrefs.get_xrefs_by_dst(0x1FFF36F4)  # the value in .bss

    assert len(timenow_cp_xrefs) == 1
    assert next(iter(timenow_cp_xrefs)) == XRef(ins_addr=0x23C9, dst=0x23D4, xref_type=XRefType.Read)
    # sys_now (2), time_isr (3) == 5
    assert len(timenow_xrefs) == 5


def test_p2im_drone_with_inits():
    bin_path = os.path.join(test_location, "armel", "p2im_drone.elf")
    proj = angr.Project(bin_path, auto_load_libs=False)
    cfg = proj.analyses.CFG(data_references=True)

    func = cfg.functions["Peripherals_Init"]
    state = proj.factory.blank_state()
    prop = proj.analyses.Propagator(func=func, base_state=state)

    init_finder = proj.analyses.InitializationFinder(func=func, replacements=prop.replacements)
    overlay_state = init_finder.overlay_state

    cfg.do_full_xrefs(overlay_state=overlay_state)

    h12c1_inst_xrefs = proj.kb.xrefs.get_xrefs_by_dst(0x20001500)
    assert len(h12c1_inst_xrefs) == 5


if __name__ == "__main__":
    test_lwip_udpecho_bm()
    test_lwip_udpecho_bm_the_better_way()
    test_p2im_drone_with_inits()



from typing import Tuple

from ...code_location import CodeLocation


class ExternalCodeLocation(CodeLocation):
    """
    Stands for a program point that originates from outside an analysis' scope.
    i.e. a value loaded from rdi in a callee where the caller has not been analyzed.
    """

    __slots__ = ("call_string",)

    def __init__(self, call_string: Tuple[int, ...] = None):
        super().__init__(0, None)
        self.call_string = call_string if call_string is not None else ()

    def __repr__(self):
        return f"[External {[hex(x) if isinstance(x, int) else x for x in self.call_string]}]"

    def __hash__(self):
        """
        returns the hash value of self.
        """
        if self._hash is None:
            self._hash = hash(
                (
                    self.block_addr,
                    self.stmt_idx,
                    self.sim_procedure,
                    self.ins_addr,
                    self.context,
                    self.block_idx,
                    self.call_string,
                )
            )
        return self._hash



import os
import claripy
import angr


def test_rcr():
    p = angr.Project(
        os.path.join(os.path.dirname(__file__), "..", "..", "binaries", "tests", "i386", "rcr_test"),
        auto_load_libs=False,
    )
    result = p.factory.successors(p.factory.entry_state()).successors[0]
    assert claripy.is_true(result.regs.cl == 8)


if __name__ == "__main__":
    test_rcr()



import angr


def test_file_unlink():
    # Initialize a blank state with an arbitrary errno location
    state = angr.SimState(arch="AMD64", mode="symbolic")
    state.libc.errno_location = 0xA0000000
    state.libc.errno = 0

    # Create a file 'test'
    fd = state.posix.open(b"test", 1)
    state.posix.close(fd)

    # Ensure 'test' was in fact created
    assert b"/test" in state.fs._files

    # Store the filename in memory
    path_addr = 0xB0000000
    state.memory.store(path_addr, b"test\x00")

    # Unlink 'test': should return 0 and leave ERRNO unchanged
    unlink = angr.SIM_PROCEDURES["posix"]["unlink"]()
    state.scratch.sim_procedure = unlink
    rval = unlink.execute(state, arguments=[path_addr]).ret_expr
    assert rval == 0
    assert state.solver.eval(state.libc.errno) == 0

    # Check that 'test' was in fact deleted
    assert state.fs._files == {}

    # Unlink again: should return -1 and set ERRNO to ENOENT
    unlink = angr.SIM_PROCEDURES["posix"]["unlink"]()
    state.scratch.sim_procedure = unlink
    rval = unlink.execute(state, arguments=[path_addr]).ret_expr
    assert rval == -1
    assert state.solver.eval(state.libc.errno) == state.posix.ENOENT



import sys
import platform
import os
import tempfile
import subprocess
import shutil

from common import has_32_bit_compiler_support

import angr


test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")

# Note: Reassembler is intensively tested by Patcherex test cases on CGC binaries.


def is_linux_x64():
    return sys.platform.startswith("linux") and platform.machine().endswith("64")


def is_linux():
    return sys.platform.startswith("linux")


def test_data_reference_collection_in_add():
    # Issue reported and test binary provided by Antonio F. Montoya
    # Fixed in https://github.com/angr/pyvex/pull/192

    p = angr.Project(os.path.join(test_location, "x86_64", "df_gcc_-O1"), auto_load_libs=False)
    vexblock_opt0 = p.factory.block(0x402431, opt_level=0).vex
    vexblock_opt1 = p.factory.block(0x402431, opt_level=1).vex
    vexblock_opt1_nostmt = p.factory.block(0x402431, opt_level=1, collect_data_refs=True).vex_nostmt

    cfg = p.analyses.CFG()

    cfg._model.memory_data = {}
    cfg._collect_data_references(vexblock_opt0, 0x402431)
    memory_data_opt0 = cfg._model.memory_data

    cfg._model.memory_data = {}
    # bypass the IRSB unoptimization step
    cfg._collect_data_references_by_scanning_stmts(vexblock_opt1, 0x402431)
    memory_data_opt1 = cfg._model.memory_data

    cfg._model.memory_data = {}
    cfg._collect_data_references(vexblock_opt1_nostmt, 0x402431)
    memory_data_opt1_nostmt = cfg._model.memory_data

    assert memory_data_opt0.keys() == memory_data_opt1.keys()
    assert memory_data_opt0.keys() == memory_data_opt1_nostmt.keys()


def test_ln_gcc_O2():
    # Issue reported and test binary provided by Antonio F. Montoya

    p = angr.Project(os.path.join(test_location, "x86_64", "ln_gcc_-O2"), auto_load_libs=False)
    r = p.analyses.Reassembler(syntax="at&t")
    r.symbolize()
    r.remove_unnecessary_stuff()
    assembly = r.assembly(comments=True, symbolized=True)

    # There should be two symbols with the same name: file_name. Reassembler renames the second one to file_name_0.
    # Test their existence.
    assert "\nfile_name:" in assembly and "\nfile_name_0:" in assembly

    if is_linux_x64():
        # we should be able to compile it and run it ... if we are running on x64 Linux
        tempdir = tempfile.mkdtemp(prefix="angr_test_reassembler_")
        asm_filename = "ln_gcc-O2.s"
        bin_filename = "ln_gcc-O2"
        asm_filepath = os.path.join(tempdir, asm_filename)
        bin_filepath = os.path.join(tempdir, bin_filename)
        with open(asm_filepath, "w", encoding="ascii") as f:
            f.write(assembly)
        # Call out to GCC, and it should return 0. Otherwise check_call() will raise an exception.
        subprocess.check_call(
            ["gcc", "-no-pie", asm_filepath, "-o", bin_filepath], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        # Run the generated binary file, and it should not crash (which is a pretty basic requirement, I know)
        subprocess.check_call([bin_filepath, "--help"], stdout=subprocess.DEVNULL)
        # Pick up after ourselves
        shutil.rmtree(tempdir)


def test_chmod_gcc_O1():
    # Issue reported and test binary provided by Antonio F. Montoya

    p = angr.Project(os.path.join(test_location, "x86_64", "chmod_gcc_-O1"), auto_load_libs=False)
    r = p.analyses.Reassembler(syntax="at&t")
    r.symbolize()
    r.remove_unnecessary_stuff()
    assembly = r.assembly(comments=True, symbolized=True)

    if is_linux_x64():
        # we should be able to compile it and run it ... if we are running on x64 Linux
        tempdir = tempfile.mkdtemp(prefix="angr_test_reassembler_")
        asm_filename = "chmod_gcc-O1.s"
        bin_filename = "chmod_gcc-O1"
        asm_filepath = os.path.join(tempdir, asm_filename)
        bin_filepath = os.path.join(tempdir, bin_filename)
        with open(asm_filepath, "w", encoding="ascii") as f:
            f.write(assembly)
        # Call out to GCC, and it should return 0. Otherwise check_call() will raise an exception.
        subprocess.check_call(
            ["gcc", "-no-pie", asm_filepath, "-o", bin_filepath], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        # Run the generated binary file, and it should not crash (which is a pretty basic requirement, I know)
        subprocess.check_call([bin_filepath, "--help"], stdout=subprocess.DEVNULL)
        # Pick up after ourselves
        shutil.rmtree(tempdir)


def test_ex_gpp():
    # Issue reported and test binary provided by Antonio F. Montoya

    p = angr.Project(os.path.join(test_location, "x86_64", "ex_g++"), auto_load_libs=False)
    r = p.analyses.Reassembler(syntax="at&t")
    r.symbolize()
    r.remove_unnecessary_stuff()
    assembly = r.assembly(comments=True, symbolized=True)

    if is_linux_x64():
        # we should be able to compile it and run it ... if we are running on x64 Linux
        tempdir = tempfile.mkdtemp(prefix="angr_test_reassembler_")
        asm_filename = "ex_g++.s"
        bin_filename = "ex_g++"
        asm_filepath = os.path.join(tempdir, asm_filename)
        bin_filepath = os.path.join(tempdir, bin_filename)
        with open(asm_filepath, "w", encoding="ascii") as f:
            f.write(assembly)
        # Call out to GCC, and it should return 0. Otherwise check_call() will raise an exception.
        subprocess.check_call(
            ["g++", "-no-pie", asm_filepath, "-o", bin_filepath], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        # Run the generated binary file and check the output
        output = subprocess.check_output([bin_filepath])
        assert output == b"A1\nA2\n"
        # Pick up after ourselves
        shutil.rmtree(tempdir)


def test_df_gcc_O1():
    # Issue reported and test binary provided by Antonio F. Montoya

    p = angr.Project(os.path.join(test_location, "x86_64", "df_gcc_-O1"), auto_load_libs=False)
    r = p.analyses.Reassembler(syntax="at&t")
    r.symbolize()
    r.remove_unnecessary_stuff()
    assembly = r.assembly(comments=True, symbolized=True)

    if is_linux_x64():
        # we should be able to compile it and run it ... if we are running on x64 Linux
        tempdir = tempfile.mkdtemp(prefix="angr_test_reassembler_")
        asm_filename = "df_gcc-O1.s"
        bin_filename = "df_gcc-O1"
        asm_filepath = os.path.join(tempdir, asm_filename)
        bin_filepath = os.path.join(tempdir, bin_filename)
        with open(asm_filepath, "w", encoding="ascii") as f:
            f.write(assembly)
        # Call out to GCC, and it should return 0. Otherwise check_call() will raise an exception.
        subprocess.check_call(
            ["gcc", "-no-pie", asm_filepath, "-o", bin_filepath], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        # Run the generated binary file, and it should not crash (which is a pretty basic requirement, I know)
        subprocess.check_call([bin_filepath, "--help"], stdout=subprocess.DEVNULL)
        # Pick up after ourselves
        shutil.rmtree(tempdir)


def test_dir_gcc_O0():
    # Issue reported and test binary provided by Antonio F. Montoya

    p = angr.Project(os.path.join(test_location, "x86_64", "dir_gcc_-O0"), auto_load_libs=False)
    r = p.analyses.Reassembler(syntax="at&t")
    r.symbolize()
    r.remove_unnecessary_stuff()
    assembly = r.assembly(comments=True, symbolized=True)

    if is_linux_x64():
        # we should be able to compile it and run it ... if we are running on x64 Linux
        tempdir = tempfile.mkdtemp(prefix="angr_test_reassembler_")
        asm_filename = "dir_gcc-O0.s"
        bin_filename = "dir_gcc-O0"
        asm_filepath = os.path.join(tempdir, asm_filename)
        bin_filepath = os.path.join(tempdir, bin_filename)
        with open(asm_filepath, "w", encoding="ascii") as f:
            f.write(assembly)
        # Call out to GCC, and it should return 0. Otherwise check_call() will raise an exception.
        subprocess.check_call(
            ["gcc", "-no-pie", asm_filepath, "-o", bin_filepath], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        # Run the generated binary file, and it should not crash (which is a pretty basic requirement, I know)
        subprocess.check_call([bin_filepath, "--help"], stdout=subprocess.DEVNULL)
        subprocess.check_call([bin_filepath, "-la", "/"], stdout=subprocess.DEVNULL)
        # Pick up after ourselves
        shutil.rmtree(tempdir)


def test_helloworld():
    # Reassembler complains about TYPE_OTHER symbols, which is because it's trying to classify bytes inside the ELF
    # header as pointers. We identify the ELF header in CFGFast to workaround this problem.
    # https://github.com/angr/angr/issues/1630

    p = angr.Project(os.path.join(test_location, "x86_64", "hello_world"), auto_load_libs=False)
    r = p.analyses.Reassembler(syntax="at&t")
    r.symbolize()
    r.remove_unnecessary_stuff()
    _ = r.assembly(comments=True, symbolized=True)

    # No exception should have been raised


def test_helloworld_gcc9():
    # New versions of GCC changed the names of init and fini sections.
    # https://github.com/angr/patcherex/issues/39

    p = angr.Project(os.path.join(test_location, "x86_64", "hello_gcc9_reassembler"), auto_load_libs=False)
    r = p.analyses.Reassembler(syntax="at&t")
    r.symbolize()
    r.remove_unnecessary_stuff()
    assembly = r.assembly(comments=True, symbolized=True)

    if is_linux_x64():
        # we should be able to compile it and run it ... if we are running on x64 Linux
        tempdir = tempfile.mkdtemp(prefix="angr_test_reassembler_")
        asm_filename = "hello.s"
        bin_filename = "hello"
        asm_filepath = os.path.join(tempdir, asm_filename)
        bin_filepath = os.path.join(tempdir, bin_filename)
        with open(asm_filepath, "w", encoding="ascii") as f:
            f.write(assembly)
        # Call out to GCC, and it should return 0. Otherwise check_call() will raise an exception.
        subprocess.check_call(
            ["gcc", "-no-pie", asm_filepath, "-o", bin_filepath], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        # Run the generated binary file, and it should not crash
        subprocess.check_call([bin_filepath], stdout=subprocess.DEVNULL)
        # Pick up after ourselves
        shutil.rmtree(tempdir)


def test_partial_pie_ls_x86():
    # https://github.com/angr/patcherex/issues/39
    # a GCC-generated X86 binary with a few functions somehow being PIE

    p = angr.Project(os.path.join(test_location, "i386", "ls_gcc_7.5_reassembler"), auto_load_libs=False)
    r = p.analyses.Reassembler(syntax="at&t")
    r.symbolize()
    r.remove_unnecessary_stuff()
    assembly = r.assembly(comments=True, symbolized=True)

    if is_linux() and has_32_bit_compiler_support():
        # we should be able to compile it and run it ... if we are running on x64 Linux
        tempdir = tempfile.mkdtemp(prefix="angr_test_reassembler_")
        asm_filename = "ls.s"
        bin_filename = "ls"
        asm_filepath = os.path.join(tempdir, asm_filename)
        bin_filepath = os.path.join(tempdir, bin_filename)
        with open(asm_filepath, "w", encoding="ascii") as f:
            f.write(assembly)
        # Call out to GCC, and it should return 0. Otherwise check_call() will raise an exception.
        subprocess.check_call(
            ["gcc", "-m32", "-no-pie", asm_filepath, "-o", bin_filepath],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        # Run the generated binary file, and it should not crash
        subprocess.check_call([bin_filepath], stdout=subprocess.DEVNULL)
        # We can also run it with "-h"
        o = subprocess.check_output([bin_filepath, "--version"])
        assert (
            o == b"ls (GNU coreutils) 8.30\n"
            b"Copyright (C) 2018 Free Software Foundation, Inc.\n"
            b"License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>.\n"
            b"This is free software: you are free to change and redistribute it.\n"
            b"There is NO WARRANTY, to the extent permitted by law.\n\n"
            b"Written by Richard M. Stallman and David MacKenzie.\n"
        )
        # Pick up after ourselves
        shutil.rmtree(tempdir)


if __name__ == "__main__":
    test_data_reference_collection_in_add()
    test_ln_gcc_O2()
    test_chmod_gcc_O1()
    test_ex_gpp()
    test_df_gcc_O1()
    test_dir_gcc_O0()
    test_helloworld()
    test_helloworld_gcc9()
    test_partial_pie_ls_x86()



# pylint: disable=missing-class-docstring,disable=no-self-use
import logging
import os
import unittest

import angr

l = logging.getLogger("angr.tests")

test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


class TestEcho(unittest.TestCase):
    def _run_echo_haha(self, arch):
        # auto_load_libs can't be disabled as the test fails
        p = angr.Project(os.path.join(test_location, arch, "echo"), use_sim_procedures=False)
        s = p.factory.full_init_state(
            mode="symbolic_approximating", args=["echo", "haha"], add_options={angr.options.STRICT_PAGE_ACCESS}
        )
        pg = p.factory.simulation_manager(s)
        pg.run(until=lambda lpg: len(lpg.active) != 1)

        assert len(pg.deadended) == 1
        assert len(pg.active) == 0
        # Need to dump by path because the program closes stdout
        assert pg.deadended[0].posix.stdout.concretize() == [b"haha\n"]

    def test_echo_haha(self):
        self._run_echo_haha("x86_64")


if __name__ == "__main__":
    unittest.main()



# pylint:disable=isinstance-second-argument-not-valid-type
import weakref
from typing import Set, Optional, Any, Tuple, Union, List, Dict, DefaultDict, TYPE_CHECKING
from collections import defaultdict
import logging
import time

import claripy
import ailment
import pyvex
from archinfo.arch_arm import is_arm_arch

from ... import sim_options
from ...knowledge_plugins.propagations import PropagationModel
from ...storage.memory_mixins import LabeledMemory
from ...errors import SimMemoryMissingError
from ...code_location import CodeLocation
from ...storage.memory_object import SimMemoryObject, SimLabeledMemoryObject
from .. import register_analysis
from ..analysis import Analysis
from ..forward_analysis import ForwardAnalysis, FunctionGraphVisitor, SingleNodeGraphVisitor
from .engine_vex import SimEnginePropagatorVEX
from .engine_ail import SimEnginePropagatorAIL
from .prop_value import PropValue, Detail

if TYPE_CHECKING:
    from archinfo import Arch
    from angr.project import Project
    from angr.analyses.reaching_definitions.reaching_definitions import ReachingDefinitionsModel


_l = logging.getLogger(name=__name__)


class PropagatorState:
    """
    Describes the base state used in Propagator.

    :ivar arch:             Architecture of the binary.
    :ivar gp:               alue of the global pointer for MIPS binaries.
    :ivar _replacements:    Stores expressions to replace, keyed by CodeLocation instances
    :ivar _equivalence:      Stores equivalence constraints that Propagator discovers during the analysis.
    :ivar _only_consts:     Only track constants.
    :ivar _expr_used_locs:  A dict keyed by expressions and valued by CodeLocations where the expression is used.
    :ivar _max_prop_expr_occurrence:    The upperbound for the number of occurrences of an expression for Propagator
                            to propagate that expression to new locations (and replace the original expression).
                            Setting it to 0 disables this limit, which means Propagator will always propagate
                            expressions regardless of how many times it has been propagated.
    """

    __slots__ = (
        "arch",
        "gpr_size",
        "_expr_used_locs",
        "_only_consts",
        "_replacements",
        "_equivalence",
        "project",
        "_store_tops",
        "_gp",
        "_max_prop_expr_occurrence",
        "__weakref__",
    )

    _tops = {}

    def __init__(
        self,
        arch: "Arch",
        project: Optional["Project"] = None,
        replacements: Optional[DefaultDict[CodeLocation, Dict]] = None,
        only_consts: bool = False,
        expr_used_locs: Optional[DefaultDict[Any, Set[CodeLocation]]] = None,
        equivalence: Optional[Set["Equivalence"]] = None,
        store_tops: bool = True,
        gp: Optional[int] = None,
        max_prop_expr_occurrence: int = 1,
    ):
        self.arch = arch
        self.gpr_size = arch.bits // arch.byte_width  # size of the general-purpose registers

        # propagation count of each expression
        self._expr_used_locs = defaultdict(set) if expr_used_locs is None else expr_used_locs
        self._only_consts = only_consts
        self._replacements = defaultdict(dict) if replacements is None else replacements
        self._equivalence: Set[Equivalence] = equivalence if equivalence is not None else set()
        self._store_tops = store_tops
        self._max_prop_expr_occurrence = max_prop_expr_occurrence

        # architecture-specific information
        self._gp: Optional[int] = gp  # Value of gp for MIPS32 and 64 binaries

        self.project = project

    def __repr__(self):
        return "<PropagatorState>"

    def _get_weakref(self):
        return weakref.proxy(self)

    @staticmethod
    def _is_const(v) -> bool:
        if isinstance(v, (int, ailment.Expr.Const)):
            return True
        if isinstance(v, claripy.ast.BV) and v.op == "BVV":
            return True
        if isinstance(v, claripy.ast.FP) and v.op == "FPV":
            return True
        if isinstance(v, claripy.ast.Bool) and v.op == "BoolV":
            return True
        return False

    @staticmethod
    def _mo_cmp(
        mo_self: Union["SimMemoryObject", "SimLabeledMemoryObject"],
        mo_other: Union["SimMemoryObject", "SimLabeledMemoryObject"],
        addr: int,
        size: int,
    ):  # pylint:disable=unused-argument
        # comparing bytes from two sets of memory objects
        # we don't need to resort to byte-level comparison. object-level is good enough.

        if mo_self.object.symbolic or mo_other.object.symbolic:
            if type(mo_self) is SimLabeledMemoryObject and type(mo_other) is SimLabeledMemoryObject:
                return mo_self.label == mo_other.label and mo_self.object is mo_other.object
            if type(mo_self) is SimMemoryObject and type(mo_other) is SimMemoryObject:
                return mo_self.object is mo_other.object
            # SimMemoryObject vs SimLabeledMemoryObject -> the label must be different
            return False
        return None

    @staticmethod
    def top(bits: int) -> claripy.ast.Bits:
        """
        Get a TOP value.

        :param size:    Width of the TOP value (in bits).
        :return:        The TOP value.
        """

        if bits in PropagatorState._tops:
            return PropagatorState._tops[bits]
        r = claripy.BVS("TOP", bits, explicit_name=True)
        PropagatorState._tops[bits] = r
        return r

    @staticmethod
    def is_top(expr) -> bool:
        """
        Check if the given expression is a TOP value.

        :param expr:    The given expression.
        :return:        True if the expression is TOP, False otherwise.
        """
        if isinstance(expr, claripy.ast.Base):
            if expr.op == "BVS" and expr.args[0] == "TOP":
                return True
            if "TOP" in expr.variables:
                return True
        return False

    def copy(self) -> "PropagatorState":
        raise NotImplementedError()

    def merge(self, *others):
        state = self.copy()
        merge_occurred = False

        for o in others:
            merge_occurred |= PropagatorAnalysis.merge_replacements(state._replacements, o._replacements)

            if state._equivalence != o._equivalence:
                merge_occurred = True
                state._equivalence |= o._equivalence

        return state, merge_occurred

    def add_replacement(self, codeloc, old, new):
        """
        Add a replacement record: Replacing expression `old` with `new` at program location `codeloc`.
        If the self._only_consts flag is set to true, only constant values will be set.

        :param CodeLocation codeloc:    The code location.
        :param old:                     The expression to be replaced.
        :param new:                     The expression to replace with.
        :return:                        None
        """
        if self.is_top(new):
            return

        if self._only_consts:
            if self._is_const(new) or self.is_top(new):
                self._replacements[codeloc][old] = new
        else:
            self._replacements[codeloc][old] = new

    def filter_replacements(self):
        pass


# VEX state


class PropagatorVEXState(PropagatorState):
    """
    Describes the state used in the VEX engine of Propagator.
    """

    __slots__ = (
        "_registers",
        "_stack_variables",
        "do_binops",
    )

    def __init__(
        self,
        arch,
        project=None,
        registers=None,
        local_variables=None,
        replacements=None,
        only_consts=False,
        expr_used_locs=None,
        do_binops=True,
        store_tops=True,
        gp=None,
        max_prop_expr_occurrence: int = 1,
    ):
        super().__init__(
            arch,
            project=project,
            replacements=replacements,
            only_consts=only_consts,
            expr_used_locs=expr_used_locs,
            store_tops=store_tops,
            gp=gp,
            max_prop_expr_occurrence=max_prop_expr_occurrence,
        )
        self.do_binops = do_binops
        self._registers = (
            LabeledMemory(memory_id="reg", top_func=self.top, page_kwargs={"mo_cmp": self._mo_cmp})
            if registers is None
            else registers
        )
        self._stack_variables = (
            LabeledMemory(memory_id="mem", top_func=self.top, page_kwargs={"mo_cmp": self._mo_cmp})
            if local_variables is None
            else local_variables
        )

        self._registers.set_state(self)
        self._stack_variables.set_state(self)

    def __repr__(self):
        return "<PropagatorVEXState>"

    def copy(self) -> "PropagatorVEXState":
        cp = PropagatorVEXState(
            self.arch,
            project=self.project,
            registers=self._registers.copy(),
            local_variables=self._stack_variables.copy(),
            replacements=self._replacements.copy(),
            expr_used_locs=self._expr_used_locs.copy(),
            only_consts=self._only_consts,
            do_binops=self.do_binops,
            store_tops=self._store_tops,
            gp=self._gp,
            max_prop_expr_occurrence=self._max_prop_expr_occurrence,
        )

        return cp

    def merge(self, *others: "PropagatorVEXState") -> Tuple["PropagatorVEXState", bool]:
        state = self.copy()
        merge_occurred = state._registers.merge([o._registers for o in others], None)
        merge_occurred |= state._stack_variables.merge([o._stack_variables for o in others], None)
        return state, merge_occurred

    def store_local_variable(self, offset, size, value, endness):  # pylint:disable=unused-argument
        # TODO: Handle size
        self._stack_variables.store(offset, value, size=size, endness=endness)

    def load_local_variable(self, offset, size, endness):  # pylint:disable=unused-argument
        # TODO: Handle size
        try:
            return self._stack_variables.load(offset, size=size, endness=endness)
        except SimMemoryMissingError:
            return self.top(size * self.arch.byte_width)

    def store_register(self, offset, size, value):
        self._registers.store(offset, value, size=size)

    def load_register(self, offset, size):
        # TODO: Fix me
        if size != self.gpr_size:
            return self.top(size * self.arch.byte_width)

        try:
            return self._registers.load(offset, size=size)
        except SimMemoryMissingError:
            return self.top(size * self.arch.byte_width)


# AIL state


class Equivalence:
    """
    Describes an equivalence relationship between two atoms.
    """

    __slots__ = (
        "codeloc",
        "atom0",
        "atom1",
    )

    def __init__(self, codeloc, atom0, atom1):
        self.codeloc = codeloc
        self.atom0 = atom0
        self.atom1 = atom1

    def __repr__(self):
        return f"<Eq@{self.codeloc!r}: {self.atom0!r}=={self.atom1!r}>"

    def __eq__(self, other):
        return (
            type(other) is Equivalence
            and other.codeloc == self.codeloc
            and other.atom0 == self.atom0
            and other.atom1 == self.atom1
        )

    def __hash__(self):
        return hash((Equivalence, self.codeloc, self.atom0, self.atom1))


class PropagatorAILState(PropagatorState):
    """
    Describes the state used in the AIL engine of Propagator.
    """

    __slots__ = (
        "_registers",
        "_stack_variables",
        "_tmps",
        "temp_expressions",
        "register_expressions",
        "last_stack_store",
        "global_stores",
        "block_initial_reg_values",
        "_sp_adjusted",
    )

    def __init__(
        self,
        arch,
        project=None,
        replacements=None,
        only_consts=False,
        expr_used_locs=None,
        equivalence=None,
        stack_variables=None,
        registers=None,
        gp=None,
        block_initial_reg_values=None,
        max_prop_expr_occurrence: int = 1,
        sp_adjusted: bool = False,
    ):
        super().__init__(
            arch,
            project=project,
            replacements=replacements,
            only_consts=only_consts,
            expr_used_locs=expr_used_locs,
            equivalence=equivalence,
            gp=gp,
            max_prop_expr_occurrence=max_prop_expr_occurrence,
        )

        self._stack_variables = (
            LabeledMemory(memory_id="mem", top_func=self.top, page_kwargs={"mo_cmp": self._mo_cmp})
            if stack_variables is None
            else stack_variables
        )
        self._registers = (
            LabeledMemory(memory_id="reg", top_func=self.top, page_kwargs={"mo_cmp": self._mo_cmp})
            if registers is None
            else registers
        )
        self._tmps = {}
        self.temp_expressions = {}
        self.register_expressions = {}
        self.block_initial_reg_values: DefaultDict[
            Tuple[int, int], List[Tuple[ailment.Expr.Register, ailment.Expr.Const]]
        ] = (defaultdict(list) if block_initial_reg_values is None else block_initial_reg_values)
        self._sp_adjusted: bool = sp_adjusted

        self._registers.set_state(self)
        self._stack_variables.set_state(self)
        # last_stack_store stores the most recent stack store statement with a non-concrete or unresolvable address. we
        # use this information to determine if stack reads after this store can be safely resolved to definitions prior
        # to the stack read.
        self.last_stack_store: Optional[Tuple[int, int, ailment.Stmt.Store]] = None
        self.global_stores: List[Tuple[int, int, Any, ailment.Stmt.Store]] = []

    def __repr__(self):
        return "<PropagatorAILState>"

    def copy(self) -> "PropagatorAILState":
        rd = PropagatorAILState(
            self.arch,
            project=self.project,
            replacements=self._replacements.copy(),
            expr_used_locs=self._expr_used_locs.copy(),
            only_consts=self._only_consts,
            equivalence=self._equivalence.copy(),
            stack_variables=self._stack_variables.copy(),
            registers=self._registers.copy(),
            block_initial_reg_values=self.block_initial_reg_values.copy(),
            # drop tmps
            gp=self._gp,
            max_prop_expr_occurrence=self._max_prop_expr_occurrence,
            sp_adjusted=self._sp_adjusted,
        )

        return rd

    @staticmethod
    def is_const_or_register(value: Optional[Union[ailment.Expr.Expression, claripy.ast.Bits]]) -> bool:
        if value is None:
            return False
        if isinstance(value, claripy.ast.BV):
            return not value.symbolic
        if isinstance(value, ailment.Expr.Register):
            return True
        if isinstance(value, ailment.Expr.Const) or (isinstance(value, int) and value == 0):
            return True
        # more hacks: also store the eq comparisons
        if isinstance(value, ailment.Expr.BinaryOp) and value.op == "CmpEQ":
            if all(isinstance(arg, (ailment.Expr.Const, ailment.Expr.Tmp)) for arg in value.operands):
                return True
        # more hacks: also store the conversions
        if isinstance(value, ailment.Expr.Convert) and PropagatorAILState.is_const_or_register(value.operand):
            return True
        return False

    def merge(self, *others) -> Tuple["PropagatorAILState", bool]:
        state, merge_occurred = super().merge(*others)
        state: "PropagatorAILState"

        merge_occurred |= state._registers.merge([o._registers for o in others], None)
        merge_occurred |= state._stack_variables.merge([o._stack_variables for o in others], None)

        return state, merge_occurred

    def store_temp(self, tmp_idx: int, value: PropValue):
        self._tmps[tmp_idx] = value

    def load_tmp(self, tmp_idx: int) -> Optional[PropValue]:
        return self._tmps.get(tmp_idx, None)

    def store_register(self, reg: ailment.Expr.Register, value: PropValue) -> None:
        if isinstance(value, ailment.Expr.Expression) and value.has_atom(reg, identity=False):
            return

        for offset, chopped_value, size, label in value.value_and_labels():
            self._registers.store(
                reg.reg_offset + offset,
                chopped_value,
                size=size,
                label=label,
                endness=self.project.arch.register_endness,
            )

    def store_stack_variable(
        self, sp_offset: int, new: PropValue, endness=None
    ) -> None:  # pylint:disable=unused-argument
        # normalize sp_offset to handle negative offsets
        sp_offset += 0x65536
        sp_offset &= (1 << self.arch.bits) - 1

        for offset, value, size, label in new.value_and_labels():
            self._stack_variables.store(sp_offset + offset, value, size=size, endness=endness, label=label)

    def load_register(self, reg: ailment.Expr.Register) -> Optional[PropValue]:
        try:
            value, labels = self._registers.load_with_labels(
                reg.reg_offset, size=reg.size, endness=self.project.arch.register_endness
            )
        except SimMemoryMissingError:
            # value does not exist
            return None

        prop_value = PropValue.from_value_and_labels(value, labels)
        return prop_value

    def load_stack_variable(self, sp_offset: int, size, endness=None) -> Optional[PropValue]:
        # normalize sp_offset to handle negative offsets
        sp_offset += 0x65536
        sp_offset &= (1 << self.arch.bits) - 1
        try:
            value, labels = self._stack_variables.load_with_labels(sp_offset, size=size, endness=endness)
        except SimMemoryMissingError as ex:
            # the stack variable does not exist - however, maybe some portion of it exists!
            if ex.missing_addr > sp_offset:
                # some data exist. load again
                try:
                    value, labels = self._stack_variables.load_with_labels(
                        sp_offset, size=ex.missing_addr - sp_offset, endness=endness
                    )
                    # then we zero-extend both the value and labels
                    if value is not None and len(labels) == 1 and labels[0][0] == 0:
                        value = claripy.ZeroExt(ex.missing_size * self.arch.byte_width, value)
                        offset, offset_in_expr, size, label = labels[0]
                        labels = ((offset, offset_in_expr, size + ex.missing_size, label),)
                except SimMemoryMissingError:
                    # failed again... welp
                    return None
            else:
                return None

        prop_value = PropValue.from_value_and_labels(value, labels)
        return prop_value

    def add_replacement(self, codeloc, old, new):
        if self._only_consts:
            if self.is_const_or_register(new) or self.is_top(new):
                pass
            else:
                new = self.top(1)

        # do not replace anything with a call expression
        if isinstance(new, ailment.statement.Call):
            return
        else:
            from .call_expr_finder import CallExprFinder  # pylint:disable=import-outside-toplevel

            callexpr_finder = CallExprFinder()
            callexpr_finder.walk_expression(new)
            if callexpr_finder.has_call:
                return

        if self.is_top(new):
            # eliminate the past propagation of this expression
            self._replacements[codeloc][old] = self.top(1)  # placeholder
            return

        # count-based propagation rule only matters when we are performing a full-function copy propagation
        if self._max_prop_expr_occurrence == 0:
            if (
                isinstance(old, ailment.Expr.Tmp)
                or isinstance(old, ailment.Expr.Register)
                and old.reg_offset in {self.arch.sp_offset, self.arch.bp_offset}
            ):
                self._replacements[codeloc][old] = new
        else:
            prop_count = 0
            if (
                not isinstance(old, ailment.Expr.Tmp)
                and isinstance(new, ailment.Expr.Expression)
                and not isinstance(new, ailment.Expr.Const)
            ):
                # FIXME: We should find the definition in the RDA result and use the definition as the key
                self._expr_used_locs[new].add(codeloc)
                prop_count = len(self._expr_used_locs[new])

            if (  # pylint:disable=too-many-boolean-expressions
                prop_count <= self._max_prop_expr_occurrence
                or isinstance(new, ailment.Expr.StackBaseOffset)
                or isinstance(new, ailment.Expr.Convert)
                and isinstance(new.operand, ailment.Expr.StackBaseOffset)
                or (
                    isinstance(old, ailment.Expr.Register)
                    and self.arch.is_artificial_register(old.reg_offset, old.size)
                )
            ):
                # we can propagate this expression
                self._replacements[codeloc][old] = new
            else:
                # eliminate the past propagation of this expression
                for codeloc_ in self._replacements:
                    if old in self._replacements[codeloc_]:
                        self._replacements[codeloc_][old] = self.top(1)

    def add_equivalence(self, codeloc, old, new):
        eq = Equivalence(codeloc, old, new)
        self._equivalence.add(eq)


class PropagatorAnalysis(ForwardAnalysis, Analysis):  # pylint:disable=abstract-method
    """
    PropagatorAnalysis implements copy propagation. It propagates values (either constant values or variables) and
    expressions inside a block or across a function.

    PropagatorAnalysis supports both VEX and AIL. The VEX propagator only performs constant propagation. The AIL
    propagator performs both constant propagation and copy propagation of depth-N expressions.

    PropagatorAnalysis performs certain arithmetic operations between constants, including but are not limited to:

    - addition
    - subtraction
    - multiplication
    - division
    - xor

    It also performs the following memory operations:

    - Loading values from a known address
    - Writing values to a stack variable
    """

    def __init__(
        self,
        func=None,
        block=None,
        func_graph=None,
        base_state=None,
        max_iterations=3,
        load_callback=None,
        stack_pointer_tracker=None,
        only_consts=False,
        completed_funcs=None,
        do_binops=True,
        store_tops=True,
        vex_cross_insn_opt=False,
        func_addr: Optional[int] = None,
        gp: Optional[int] = None,
        cache_results: bool = False,
        key_prefix: Optional[str] = None,
        reaching_definitions: Optional["ReachingDefinitionsModel"] = None,
        profiling: bool = False,
    ):
        if block is None and func is not None:
            # only func is specified. traversing a function
            self.flavor = "function"
        elif block is not None:
            # traversing a block (but func might be specified at the same time to provide extra information, e.g., the
            # value for register t9 for MIPS32/64 binaries)
            self.flavor = "block"
        else:
            raise ValueError("Unsupported analysis target.")

        start = time.perf_counter_ns() / 1000000

        self._base_state = base_state
        self._function = func
        self._func_addr = func_addr if func_addr is not None else (None if func is None else func.addr)
        self._block = block
        self._block_addr = block.addr if block is not None else None
        self._max_iterations = max_iterations
        self._load_callback = load_callback
        self._stack_pointer_tracker = stack_pointer_tracker  # only used when analyzing AIL functions
        self._only_consts = only_consts
        self._completed_funcs = completed_funcs
        self._do_binops = do_binops
        self._store_tops = store_tops
        self._vex_cross_insn_opt = vex_cross_insn_opt
        self._gp = gp
        self._prop_key_prefix = key_prefix
        self._cache_results = cache_results
        self._reaching_definitions = reaching_definitions
        self._initial_codeloc: CodeLocation
        if self.flavor == "function":
            self._initial_codeloc = CodeLocation(self._func_addr, stmt_idx=0, ins_addr=self._func_addr)
        else:  # flavor == "block"
            self._initial_codeloc = CodeLocation(self._block_addr, stmt_idx=0, ins_addr=self._block_addr)

        self.model: PropagationModel = None

        if self._cache_results:
            # Resume the analysis from the previously unfinished result
            self.model = self.kb.propagations.get(self.prop_key, None)

        if self.model is None:
            self.model = PropagationModel(
                self.prop_key,
            )
            cache_used = False
        else:
            cache_used = True

        graph_visitor: Union[SingleNodeGraphVisitor, FunctionGraphVisitor]
        if self.flavor == "block":
            graph_visitor = None
            if self._cache_results:
                graph_visitor: Optional[SingleNodeGraphVisitor] = self.model.graph_visitor

            if graph_visitor is None:
                graph_visitor = SingleNodeGraphVisitor(block)

        elif self.flavor == "function":
            graph_visitor = None
            if self._cache_results:
                graph_visitor: Optional[FunctionGraphVisitor] = self.model.graph_visitor
                if graph_visitor is not None:
                    # resume
                    resumed = graph_visitor.resume_with_new_graph(func_graph if func_graph is not None else func.graph)
                    if not resumed:
                        # clean up...
                        self.model = PropagationModel(self.prop_key)

            if graph_visitor is None:
                graph_visitor = FunctionGraphVisitor(func, func_graph)
                self.model.graph_visitor = graph_visitor

        else:
            raise TypeError(f"Unsupported flavor {self.flavor}")

        ForwardAnalysis.__init__(
            self, order_jobs=True, allow_merging=True, allow_widening=False, graph_visitor=graph_visitor
        )

        self._engine_vex = SimEnginePropagatorVEX(
            project=self.project, arch=self.project.arch, reaching_definitions=self._reaching_definitions
        )
        self._engine_ail = SimEnginePropagatorAIL(
            arch=self.project.arch,
            stack_pointer_tracker=self._stack_pointer_tracker,
            # We only propagate tmps within the same block. This is because the lifetime of tmps is one block only.
            propagate_tmps=block is not None,
            reaching_definitions=self._reaching_definitions,
        )

        # optimization: skip state copying for the initial state
        self._initial_state = None

        # performance counters
        self._analyzed_states: int = 0
        self._analyzed_statements: int = 0

        self._analyze()

        if self._cache_results:
            # update the cache
            self.kb.propagations.update(self.prop_key, self.model)

        if profiling:
            elapsed = time.perf_counter_ns() / 1000000 - start
            if self.flavor == "function":
                _l.warning("%r:", self._function)
            else:
                _l.warning("%r:", self._block)
            _l.warning("  Time elapsed: %s milliseconds", elapsed)
            _l.warning("  Cache used: %s", cache_used)
            _l.warning("  Analyzed states: %d", self._analyzed_states)
            _l.warning("  Analyzed statements: %d", self._analyzed_statements)

    @property
    def prop_key(self) -> Tuple[Optional[str], str, int, bool, bool, bool]:
        """
        Gets a key that represents the function and the "flavor" of the propagation result.
        """
        addr = self._func_addr if self._func_addr is not None else self._block_addr
        return self._prop_key_prefix, self.flavor, addr, self._do_binops, self._only_consts, self._vex_cross_insn_opt

    @property
    def replacements(self):
        return self.model.replacements

    @replacements.setter
    def replacements(self, v):
        self.model.replacements = v

    #
    # Main analysis routines
    #

    def _node_key(self, node: Union[ailment.Block, pyvex.IRSB]) -> Any:
        if type(node) is ailment.Block:
            return node.addr, node.idx
        elif type(node) is pyvex.IRSB:
            return node.addr
        # fallback
        return node

    def _pre_analysis(self):
        pass

    def _pre_job_handling(self, job):
        pass

    def _initial_abstract_state(self, node):
        if isinstance(node, ailment.Block):
            # AIL
            state = PropagatorAILState(
                self.project.arch,
                project=self.project,
                only_consts=self._only_consts,
                gp=self._gp,
                max_prop_expr_occurrence=1 if self.flavor == "function" else 0,
            )
            ail = True
            spoffset_var = ailment.Expr.StackBaseOffset(None, self.project.arch.bits, 0)
            sp_value = PropValue(
                claripy.BVV(0x7FFF_FF00, self.project.arch.bits),
                offset_and_details={0: Detail(self.project.arch.bytes, spoffset_var, self._initial_codeloc)},
            )
            state.store_register(
                ailment.Expr.Register(None, None, self.project.arch.sp_offset, self.project.arch.bits),
                sp_value,
            )
        else:
            # VEX
            state = PropagatorVEXState(
                self.project.arch,
                project=self.project,
                only_consts=self._only_consts,
                do_binops=self._do_binops,
                store_tops=self._store_tops,
                gp=self._gp,
                max_prop_expr_occurrence=1 if self.flavor == "function" else 0,
            )
            spoffset_var = self._engine_vex.sp_offset(0)
            ail = False
            state.store_register(
                self.project.arch.sp_offset,
                self.project.arch.bytes,
                spoffset_var,
            )

        if self.project.arch.name == "MIPS64":
            if self._func_addr is not None:
                if ail:
                    reg_expr = ailment.Expr.Register(
                        None, None, self.project.arch.registers["t9"][0], self.project.arch.registers["t9"][1]
                    )
                    reg_value = ailment.Expr.Const(None, None, self._func_addr, 64)
                    state.store_register(
                        reg_expr,
                        PropValue(
                            claripy.BVV(self._func_addr, 64),
                            offset_and_details={0: Detail(8, reg_value, self._initial_codeloc)},
                        ),
                    )
                else:
                    state.store_register(  # pylint:disable=too-many-function-args
                        self.project.arch.registers["t9"][0],
                        self.project.arch.registers["t9"][1],
                        claripy.BVV(self._func_addr, 64),
                    )
        elif self.project.arch.name == "MIPS32":
            if self._func_addr is not None:
                if ail:
                    reg_expr = ailment.Expr.Register(
                        None, None, self.project.arch.registers["t9"][0], self.project.arch.registers["t9"][1]
                    )
                    reg_value = ailment.Expr.Const(None, None, self._func_addr, 32)
                    state.store_register(
                        reg_expr,
                        PropValue(
                            claripy.BVV(self._func_addr, 32),
                            offset_and_details={0: Detail(4, reg_value, self._initial_codeloc)},
                        ),
                    )
                else:
                    state.store_register(  # pylint:disable=too-many-function-args
                        self.project.arch.registers["t9"][0],
                        self.project.arch.registers["t9"][1],
                        claripy.BVV(self._func_addr, 32),
                    )
        elif is_arm_arch(self.project.arch):
            if ail:
                # clear fpscr
                reg_expr = ailment.Expr.Register(None, None, *self.project.arch.registers["fpscr"])
                reg_value = ailment.Expr.Const(None, None, 0, 32)
                state.store_register(
                    reg_expr,
                    PropValue(claripy.BVV(0, 32), offset_and_details={0: Detail(4, reg_value, self._initial_codeloc)}),
                )
            else:
                state.store_register(  # pylint:disable=too-many-function-args
                    self.project.arch.registers["fpscr"][0],
                    self.project.arch.registers["fpscr"][1],
                    claripy.BVV(0, 32),
                )

        self._initial_state = state
        return state

    def _merge_states(self, node, *states: PropagatorState):
        merged_state, merge_occurred = states[0].merge(*states[1:])
        return merged_state, not merge_occurred

    def _run_on_node(self, node, state):
        self._analyzed_states += 1

        if isinstance(node, ailment.Block):
            block = node
            block_key = (node.addr, node.idx)
            engine = self._engine_ail
        else:
            block = self.project.factory.block(
                node.addr, node.size, opt_level=1, cross_insn_opt=self._vex_cross_insn_opt
            )
            block_key = node.addr
            engine = self._engine_vex
            if block.size == 0:
                # maybe the block is not decodeable
                return False, state

        if state is not self._initial_state:
            # make a copy of the state if it's not the initial state
            state = state.copy()
            state._equivalence.clear()
        else:
            # clear self._initial_state so that we *do not* run this optimization again!
            self._initial_state = None

        # Suppress spurious output
        if self._base_state is not None:
            self._base_state.options.add(sim_options.SYMBOL_FILL_UNCONSTRAINED_REGISTERS)
            self._base_state.options.add(sim_options.SYMBOL_FILL_UNCONSTRAINED_MEMORY)
        state = engine.process(
            state,
            block=block,
            project=self.project,
            base_state=self._base_state,
            load_callback=self._load_callback,
            fail_fast=self._fail_fast,
        )
        state.filter_replacements()

        self.model.node_iterations[block_key] += 1
        self.model.states[block_key] = state
        if isinstance(state, PropagatorAILState):
            self.model.block_initial_reg_values.update(state.block_initial_reg_values)

        if self.model.replacements is None:
            self.model.replacements = state._replacements
        else:
            self.merge_replacements(self.model.replacements, state._replacements)

        self.model.equivalence |= state._equivalence

        # TODO: Clear registers according to calling conventions

        if self.model.node_iterations[block_key] < self._max_iterations:
            return True, state
        else:
            return False, state

    def _process_input_state_for_successor(
        self, node, successor, input_state: Union[PropagatorAILState, PropagatorVEXState]
    ):
        if self._only_consts and isinstance(input_state, PropagatorAILState):
            key = node.addr, successor.addr
            if key in self.model.block_initial_reg_values:
                input_state: PropagatorAILState = input_state.copy()
                for reg_atom, reg_value in self.model.block_initial_reg_values[key]:
                    input_state.store_register(
                        reg_atom,
                        PropValue(
                            claripy.BVV(reg_value.value, reg_value.bits),
                            offset_and_details={0: Detail(reg_atom.size, reg_value, self._initial_codeloc)},
                        ),
                    )
                return input_state
        return input_state

    def _intra_analysis(self):
        pass

    def _check_func_complete(self, func):
        """
        Checks if a function is completely created by the CFG. Completed
        functions are passed to the Propagator at initialization. Defaults to
        being empty if no pass is initiated.

        :param func:    Function to check (knowledge_plugins.functions.function.Function)
        :return:        Bool
        """
        complete = False
        if self._completed_funcs is None:
            return complete

        if func.addr in self._completed_funcs:
            complete = True

        return complete

    def _post_analysis(self):
        """
        Post Analysis of Propagation().
        We add the current propagation replacements result to the kb if the
        function has already been completed in cfg creation.
        """

        # Filter replacements and remove all TOP values
        if self.model.replacements is not None:
            for codeloc in list(self.model.replacements.keys()):
                filtered_rep = {}
                for k, v in self.model.replacements[codeloc].items():
                    if isinstance(v, claripy.ast.Base):
                        # claripy expressions
                        if not PropagatorState.is_top(v):
                            filtered_rep[k] = v
                    else:
                        # AIL expressions
                        if not PropagatorAILState.is_top(v):
                            filtered_rep[k] = v
                self.model.replacements[codeloc] = filtered_rep

        if self._cache_results:
            self.kb.propagations.update(self.prop_key, self.model)

    def _analyze(self):
        """
        The main analysis for Propagator. Overwritten to include an optimization to stop
        analysis if we have already analyzed the entire function once.
        """
        self._pre_analysis()

        # normal analysis execution
        if self._graph_visitor is None:
            # There is no base graph that we can rely on. The analysis itself should generate successors for the
            # current job.
            # An example is the CFG recovery.

            self._analysis_core_baremetal()

        else:
            # We have a base graph to follow. Just handle the current job.

            self._analysis_core_graph()

        self._post_analysis()

    @staticmethod
    def merge_replacements(replacements_0, replacements_1) -> bool:
        """
        The replacement merging logic is special: replacements_1 is the newer replacement result and replacement_0 is
        the older result waiting to be updated. When both replacements_1 and replacement_0 have a non-top value for the
        same variable and code location, we will update the slot in replacement_0 with the value from replacement_1.

        :return:            Whether merging has happened or not.
        """
        merge_occurred = False
        for loc, vars_ in replacements_1.items():
            if loc not in replacements_0:
                replacements_0[loc] = vars_.copy()
                merge_occurred = True
            else:
                for var, repl in vars_.items():
                    if var not in replacements_0[loc]:
                        replacements_0[loc][var] = repl
                        merge_occurred = True
                    else:
                        if PropagatorState.is_top(repl) or PropagatorState.is_top(replacements_0[loc][var]):
                            t = PropagatorState.top(repl.bits if isinstance(repl, ailment.Expression) else repl.size())
                            replacements_0[loc][var] = t
                            merge_occurred = True
                        elif (
                            isinstance(replacements_0[loc][var], claripy.ast.Base) or isinstance(repl, claripy.ast.Base)
                        ) and replacements_0[loc][var] is not repl:
                            replacements_0[loc][var] = repl
                            merge_occurred = True
                        elif (
                            not isinstance(replacements_0[loc][var], claripy.ast.Base)
                            and not isinstance(repl, claripy.ast.Base)
                            and replacements_0[loc][var] != repl
                        ):
                            replacements_0[loc][var] = repl
                            merge_occurred = True
        return merge_occurred


register_analysis(PropagatorAnalysis, "Propagator")



import angr

import os

location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


def test_kb_plugins_dwarf():
    p = angr.Project(
        os.path.join(location, "x86_64", "state_merge_0"),
        load_options={
            "load_debug_info": True,
            "auto_load_libs": False,
        },
    )
    assert isinstance(p.kb.variables, angr.knowledge_plugins.VariableManager)
    p.kb.variables.load_from_dwarf()

    ret = p.kb.variables.global_manager.get_global_variables(6295620)
    assert len(ret) == 1

    v = ret.pop()
    assert v.name == "buf"


if __name__ == "__main__":
    test_kb_plugins_dwarf()



import angr

import os

test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


def test_i386():
    after_puts = 0x40105B
    else_paths = [0x401062, 0x401009]

    p = angr.Project(os.path.join(test_location, "i386", "simple_windows.exe"), auto_load_libs=False)

    s = p.factory.entry_state(args=("simple_windows.exe", "angr_can_windows?", "1497715489"))
    simgr = p.factory.simulation_manager(s)
    simgr.explore(find=after_puts, avoid=else_paths, num_find=10)

    assert len(simgr.avoid) == 0
    assert len(simgr.found) > 0
    for f in simgr.found:
        assert b"ok" in f.posix.dumps(1)


if __name__ == "__main__":
    import logging

    logging.getLogger("angr.engines").setLevel("INFO")
    test_i386()



from collections import OrderedDict

import angr
import claripy
import ctypes

from archinfo import Endness
from angr import SimState
from angr.sim_type import register_types, parse_types, SimStruct, SimTypeNumOffset


def test_simple_concrete():
    s = SimState(arch="AMD64")
    addr = 0xBA5E0

    def check_read(val):
        assert s.solver.eval(s.memory.load(addr, 8, endness=Endness.LE), cast_to=int) == val

        assert s.mem[addr].char.concrete == chr(val & 0xFF).encode()
        assert s.mem[addr].byte.concrete == val & 0xFF

        assert s.mem[addr].int16_t.concrete == ctypes.c_int16(val & 0xFFFF).value
        assert s.mem[addr].uint16_t.concrete == val & 0xFFFF

        assert s.mem[addr].qword.concrete == val

    s.memory.store(addr, claripy.BVV(0x11223344AABBCC7D, 64), endness=Endness.LE)
    check_read(0x11223344AABBCC7D)

    # test storing
    s.mem[addr].uint16_t = 0xEF6D
    check_read(0x11223344AABBEF6D)


def test_string_concrete():
    s = SimState(arch="AMD64")
    addr = 0xBA5E0

    def check_read(val):
        assert s.solver.eval(s.memory.load(addr, len(val)), cast_to=bytes) == val
        assert s.solver.eval(s.memory.load(addr + len(val), 1), cast_to=int) == 0

        assert s.mem[addr].string.concrete == val

    s.memory.store(addr, b"a string!\0")
    check_read(b"a string!")

    # not supported yet
    # s.mem[addr].string = "shorter"
    # check_read(b"shorter")

    # s.mem[addr].string = "a longer string"
    # check_read(b"a longer string")


def test_array_concrete():
    s = SimState(arch="AMD64")
    addr = 0xBA5E0

    s.memory.store(addr, claripy.BVV(0x1, 32), endness=Endness.LE)
    s.memory.store(addr + 4, claripy.BVV(0x2, 32), endness=Endness.LE)
    s.memory.store(addr + 8, claripy.BVV(0x3, 32), endness=Endness.LE)
    s.memory.store(addr + 12, claripy.BVV(0x4, 32), endness=Endness.LE)
    s.memory.store(addr + 16, claripy.BVV(0x5, 32), endness=Endness.LE)

    assert s.mem[addr].dword.array(5).concrete == [0x1, 0x2, 0x3, 0x4, 0x5]
    assert s.mem[addr].dword.array(5)[2].concrete == 0x3
    assert s.mem[addr].qword.array(2).concrete == [
        0x0000000200000001,
        0x0000000400000003,
    ]
    assert s.mem[addr].dword.array(2).array(2).concrete == [[0x1, 0x2], [0x3, 0x4]]

    s.mem[addr].dword.array(5)[3] = 10
    assert s.solver.eval(s.memory.load(addr + 12, 4, endness=Endness.LE), cast_to=int) == 10

    s.mem[addr].dword.array(5).store([20, 2, 3, 4, 5])
    assert s.mem[addr].dword.array(4).concrete == [20, 2, 3, 4]

    s.mem[addr].dword.array(2).array(2).store([[1, 2], [4, 3]])
    assert s.mem[addr].dword.array(4).concrete == [1, 2, 4, 3]


def test_pointer_concrete():
    s = SimState(arch="AMD64")
    addr = 0xBA5E0
    ptraddr = 0xCD0

    s.memory.store(ptraddr, claripy.BVV(addr, 64), endness=Endness.LE)
    s.memory.store(addr, b"abcdef\0")

    assert s.mem[ptraddr].deref.string.concrete == b"abcdef"
    s.mem[ptraddr].deref.dword = 123954
    assert s.solver.eval(s.memory.load(addr, 4, endness=Endness.LE), cast_to=int) == 123954
    assert s.mem[ptraddr].deref.dword.concrete == 123954


def test_structs():
    s = SimState(arch="AMD64")

    register_types(
        parse_types(
            """
struct test_structs {
  int a;
  long b;
};
"""
        )
    )

    s.memory.store(0x8000, bytes(16))
    s.mem[0x8000].struct.test_structs = {"a": 10, "b": 20}
    assert s.mem[0x8000].struct.test_structs.a.concrete == 10
    assert s.solver.eval(s.memory.load(0x8000, 16), cast_to=bytes) == bytes.fromhex("0a000000000000001400000000000000")


def test_struct_bitfield_simple():
    """
    Tests if a struct with bitfields like
    struct {
        uint32_t a:8, b:1, c:23;
    }
    can be used with a memview
    :return:
    """
    state = SimState(arch="AMD64")
    register_types(
        SimStruct(
            name="bitfield_struct",
            pack=True,
            fields=OrderedDict(
                [
                    ("a", SimTypeNumOffset(8, signed=False)),
                    ("b", SimTypeNumOffset(1, signed=False)),
                    ("c", SimTypeNumOffset(23, signed=False)),
                ]
            ),
        )
    )

    data = [
        (b"\x0e\x02\x00\x00", (14, 0, 1)),
        (b"\x14T\x00\x00", (20, 0, 42)),
        (b"\x04\n\x01\x00", (4, 0, 133)),
        (b"\x04j\x01\x00", (4, 0, 181)),
        (b"\x04\xa2\x01\x00", (4, 0, 209)),
        (b"\x04\xf4\x01\x00", (4, 0, 250)),
        (b"\x04\\\x02\x00", (4, 0, 302)),
        (b"\x04\x98\x02\x00", (4, 0, 332)),
        (b"\x04\xe0\x02\x00", (4, 0, 368)),
        (b"\x04\x1e\x03\x00", (4, 0, 399)),
    ]
    state.memory.store(
        0x8000,
        b"\x0e\x02\x00\x00"
        b"\x14T\x00\x00"
        b"\x04\n\x01\x00"
        b"\x04j\x01\x00"
        b"\x04\xa2\x01\x00"
        b"\x04\xf4\x01\x00"
        b"\x04\\\x02\x00"
        b"\x04\x98\x02\x00"
        b"\x04\xe0\x02\x00"
        b"\x04\x1e\x03\x00",
    )
    view = state.mem[0x8000].struct.bitfield_struct.array(5)
    for idx, (b, result) in enumerate(data):
        v = view[idx]
        s = v.concrete
        assert s.a == result[0], f"Field a was {s.a}, expected {result[0]}, from bytes {b}"
        assert v.a.concrete == result[0], f"Field a was {v.a.concrete}, expected {result[0]}, from bytes {b}"

        assert s.b == result[1], f"Field b was {s.b}, expected {result[1]}, from bytes {b}"
        assert v.b.concrete == result[1], f"Field b was {s.b}, expected {result[1]}, from bytes {b}"

        assert s.c == result[2], f"Field c was {s.c}, expected {result[2]}, from bytes {b}"
        assert v.c.concrete == result[2], f"Field c was {v.c.concrete}, expected {result[2]}, from bytes {b}"


def test_struct_bitfield_complex():
    bitfield_struct2 = angr.types.parse_type(
        """struct bitfield_struct2
    {
        uint64_t    target    : 36,
                    high8     :  8,
                    reserved  :  7,
                    next      : 12,
                    bind      :  1;
    }"""
    )

    angr.types.register_types(bitfield_struct2)
    state = SimState(arch="AMD64")
    state.memory.store(0x1000, b"\xb3\xc7\xe9|\xad\xd7\xee$")  # store some random data
    struct = state.mem[0x1000].struct.bitfield_struct2.concrete
    assert struct.target == 0xD7CE9C7B3
    assert struct.high8 == 0x7A
    assert struct.next == 0x49D
    assert struct.bind == 0
    pass


if __name__ == "__main__":
    test_simple_concrete()
    test_string_concrete()
    test_array_concrete()
    test_pointer_concrete()
    test_structs()
    test_struct_bitfield_simple()
    test_struct_bitfield_complex()



import os

import angr

test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


def test_fauxware():
    bin_path = os.path.join(test_location, "x86_64", "fauxware")
    proj = angr.Project(bin_path, auto_load_libs=False)

    cfg = proj.analyses.CFG(data_references=True, cross_references=True, normalize=True)
    func = cfg.kb.functions["main"]

    proj.analyses.Proximity(func, cfg.model, cfg.kb.xrefs)

    # once we have decompiled code, things are different...
    dec = proj.analyses.Decompiler(func, cfg=cfg.model)
    proj.analyses.Proximity(func, cfg.model, cfg.kb.xrefs, decompilation=dec)


if __name__ == "__main__":
    test_fauxware()



import os
import angr

test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


def test_vtable_extraction_x86_64():
    p = angr.Project(os.path.join(test_location, "x86_64", "cpp_classes"), auto_load_libs=False)
    vtables_sizes = {0x403CB0: 24, 0x403CD8: 16, 0x403CF8: 16, 0x403D18: 16}
    vtable_analysis = p.analyses.VtableFinder()
    vtables = vtable_analysis.vtables_list

    assert len(vtables) == 4

    for vtable in vtables:
        assert vtable.vaddr in [0x403CB0, 0x403CD8, 0x403CF8, 0x403D18]
        assert vtables_sizes[vtable.vaddr] == vtable.size


if __name__ == "__main__":
    test_vtable_extraction_x86_64()



import os
import sys
import logging

import angr

import psutil
from common import bin_location


def test_memory_watcher():
    binary = os.path.join(bin_location, "tests", "x86_64", "veritesting_a")
    proj = angr.Project(binary, auto_load_libs=False)
    simgr = proj.factory.simulation_manager()

    memory_watcher = angr.exploration_techniques.MemoryWatcher()
    simgr.use_technique(memory_watcher)

    # Initially build some paths
    while len(simgr.active) < 32 and simgr.active != []:
        simgr.step()

    # Something else went wrong..
    assert simgr.active != []

    # Set fake that memory watcher believes we're too low on memory
    memory_watcher.min_memory = psutil.virtual_memory().total

    previous_active = len(simgr.active)

    # Step once to move things over
    simgr.step()

    assert simgr.active == []
    assert len(getattr(simgr, memory_watcher.memory_stash)) == previous_active


def run_all():
    functions = globals()
    all_functions = dict(filter((lambda kv: kv[0].startswith("test_")), functions.items()))
    for f in sorted(all_functions.keys()):
        if hasattr(all_functions[f], "__call__"):
            all_functions[f]()


if __name__ == "__main__":
    logging.getLogger("angr.exploration_techniques.memory_watcher").setLevel("DEBUG")
    if len(sys.argv) > 1:
        globals()["test_" + sys.argv[1]]()
    else:
        run_all()



from typing import List, Type

from .a_div_const_add_a_mul_n_div_const import ADivConstAddAMulNDivConst
from .a_mul_const_div_shr_const import AMulConstDivShrConst
from .a_shl_const_sub_a import AShlConstSubA
from .a_sub_a_div import ASubADiv
from .a_sub_a_div_const_mul_const import ASubADivConstMulConst
from .arm_cmpf import ARMCmpF
from .bswap import Bswap
from .coalesce_same_cascading_ifs import CoalesceSameCascadingIfs
from .constant_derefs import ConstantDereferences
from .conv_const_mull_a_shift import ConvConstMullAShift
from .extended_byte_and_mask import ExtendedByteAndMask
from .remove_empty_if_body import RemoveEmptyIfBody
from .remove_redundant_ite_branch import RemoveRedundantITEBranches
from .single_bit_xor import SingleBitXor
from .a_sub_a_sub_n import ASubASubN
from .conv_a_sub0_shr_and import ConvASub0ShrAnd
from .eager_eval import EagerEvaluation
from .one_sub_bool import OneSubBool
from .bool_expr_xor_1 import BoolExprXor1
from .bitwise_or_to_logical_or import BitwiseOrToLogicalOr
from .remove_redundant_bitmasks import RemoveRedundantBitmasks
from .remove_redundant_nots import RemoveRedundantNots
from .remove_redundant_reinterprets import RemoveRedundantReinterprets
from .remove_redundant_shifts import RemoveRedundantShifts
from .remove_redundant_shifts_around_comparators import RemoveRedundantShiftsAroundComparators
from .simplify_pc_relative_loads import SimplifyPcRelativeLoads
from .basepointeroffset_add_n import BasePointerOffsetAddN
from .basepointeroffset_and_mask import BasePointerOffsetAndMask
from .remove_redundant_conversions import RemoveRedundantConversions
from .remove_cascading_conversions import RemoveCascadingConversions
from .conv_shl_shr import ConvShlShr
from .rewrite_mips_gp_loads import RewriteMipsGpLoads
from .remove_noop_conversions import RemoveNoopConversions
from .rewrite_bit_extractions import RewriteBitExtractions
from .remove_redundant_ite_comparisons import RemoveRedundantITEComparisons

from .base import PeepholeOptimizationExprBase, PeepholeOptimizationStmtBase


STMT_OPTS: List[Type[PeepholeOptimizationStmtBase]] = []
EXPR_OPTS: List[Type[PeepholeOptimizationExprBase]] = []

_g = globals().copy()
for v in _g.values():
    if isinstance(v, type) and issubclass(v, PeepholeOptimizationExprBase) and v is not PeepholeOptimizationExprBase:
        EXPR_OPTS.append(v)

    if isinstance(v, type) and issubclass(v, PeepholeOptimizationStmtBase) and v is not PeepholeOptimizationStmtBase:
        STMT_OPTS.append(v)

_g = None



import angr

import logging

l = logging.getLogger("angr_tests.test_proxy")

import os

location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


def test_tech_builder():
    class Foo:
        @staticmethod
        def setup(*args, **kwargs):
            l.debug("setup() triggered!")

        @staticmethod
        def step_state(*args, **kwargs):
            l.debug("step_state() triggered!")
            return None

        @staticmethod
        def step(simgr, stash, *args, **kwargs):
            l.debug("step() triggered!")
            return simgr.step(stash=stash, **kwargs)

        @staticmethod
        def filter(*args, **kwargs):
            l.debug("filter() triggered!")
            return None

        @staticmethod
        def complete(*args, **kwargs):
            l.debug("complete() triggered!")
            return True

    p = angr.Project(os.path.join(location, "x86_64", "fauxware"), load_options={"auto_load_libs": False})

    foo = Foo()
    proxy_tech = angr.exploration_techniques.TechniqueBuilder(
        setup=foo.setup,
        step_state=foo.step_state,
        step=foo.step,
        filter=foo.filter,
        complete=foo.complete,
    )

    pg = p.factory.simulation_manager()
    pg.use_technique(proxy_tech)
    pg.run()


if __name__ == "__main__":
    test_tech_builder()



from typing import Dict

import archinfo

import angr
from angr.sim_type import (
    SimTypeFunction,
    SimTypeInt,
    SimTypePointer,
    SimTypeChar,
    SimStruct,
    SimTypeFloat,
    SimUnion,
    SimTypeDouble,
    SimTypeLongLong,
    SimTypeLong,
    SimTypeNum,
    SimTypeReference,
    SimTypeBottom,
    SimTypeString,
)
from angr.utils.library import convert_cproto_to_py, convert_cppproto_to_py


def test_cproto_conversion():
    # A normal function declaration
    cproto_0 = "int main(int argc, char** argv);"
    pyproto_name, pyproto, the_str = convert_cproto_to_py(cproto_0)

    assert pyproto_name == "main"
    assert isinstance(pyproto, SimTypeFunction)
    assert isinstance(pyproto.args[0], SimTypeInt)
    assert isinstance(pyproto.args[1], SimTypePointer)
    assert isinstance(pyproto.args[1].pts_to.pts_to, SimTypeChar)
    assert isinstance(pyproto.returnty, SimTypeInt)

    # Directly comparing the strings... how bad can I be?
    assert the_str == (
        '# int main(int argc, char** argv);\n"main": SimTypeFunction([SimTypeInt(signed=True), SimTypePointer('
        'SimTypePointer(SimTypeChar(), offset=0), offset=0)], SimTypeInt(signed=True), arg_names=["argc", "argv"]),'
    )

    # A bad function declaration
    cproto_1 = "int bad(xxxxxxx);"
    pyproto_name, pyproto, the_str = convert_cproto_to_py(cproto_1)  # pylint:disable=unused-variable

    assert pyproto_name == "bad"
    assert pyproto is not None

    # A even worse function declaration
    # Special thanks to @schieb, see GitHub PR #958
    cproto_2 = "__attribute__ ((something)) void foo(void);"
    pyproto_name, pyproto, the_str = convert_cproto_to_py(cproto_2)  # pylint:disable=unused-variable

    assert pyproto_name == "foo"


def test_cppproto_conversion():
    # a demangled class constructor prototype, without parameter names
    proto_0 = (
        "std::basic_ifstream<char, std::char_traits<char>>::{ctor}(std::__cxx11::basic_string<char, "
        "std::char_traits<char>, std::allocator<char>> const&, std::_Ios_Openmode)"
    )
    name, proto, s = convert_cppproto_to_py(proto_0, with_param_names=False)
    assert proto.ctor is True
    assert name == "std::basic_ifstream::__ctor__"
    assert len(proto.args) == 3
    assert isinstance(proto.args[0], SimTypePointer)  # this
    assert isinstance(proto.args[1], SimTypeReference)
    assert isinstance(proto.args[1].refs, SimTypeString)
    assert proto.args[1].refs.name == "std::__cxx11::basic_string"
    assert proto.args[1].refs.unqualified_name(lang="c++") == "basic_string"

    proto_1 = "void std::basic_string<CharT,Traits,Allocator>::push_back(CharT ch)"
    name, proto, s = convert_cppproto_to_py(proto_1, with_param_names=True)
    assert name == "std::basic_string::push_back"
    assert isinstance(proto.returnty, SimTypeBottom)
    assert isinstance(proto.args[0], SimTypePointer)  # this
    assert isinstance(proto.args[1], SimTypeChar)

    proto_2 = "void std::basic_string<CharT,Traits,Allocator>::swap(basic_string& other)"
    name, proto, s = convert_cppproto_to_py(proto_2, with_param_names=True)
    assert name == "std::basic_string::swap"
    assert isinstance(proto.returnty, SimTypeBottom)
    assert isinstance(proto.args[0], SimTypePointer)  # this
    assert isinstance(proto.args[1], SimTypeReference)
    assert isinstance(proto.args[1].refs, SimTypeString)

    proto_3 = "std::ios_base::{base dtor}()"
    name, proto, s = convert_cppproto_to_py(proto_3, with_param_names=True)
    assert name == "std::ios_base::__base_dtor__"
    assert proto.dtor is True
    assert isinstance(proto.returnty, SimTypeBottom)

    proto_4 = "std::ios_base::{base dtor}()"
    name, proto, s = convert_cppproto_to_py(proto_4, with_param_names=True)
    assert name == "std::ios_base::__base_dtor__"

    proto_5 = "void foo(int & bar);"
    name, proto, s = convert_cppproto_to_py(proto_5, with_param_names=True)
    assert name == "foo"
    # note that there is no "this" pointer
    assert isinstance(proto.args[0], SimTypeReference)
    assert isinstance(proto.args[0].refs, SimTypeInt)
    assert isinstance(proto.returnty, SimTypeBottom)


def test_struct_deduplication():
    angr.types.register_types(angr.types.parse_type("struct ahdr { int a ;}"))
    angr.types.register_types(angr.types.parse_type("struct bhdr { int b ;}"))
    angr.types.register_types(angr.types.parse_type("struct chdr { int c ;}"))
    dhdr = angr.types.parse_type("struct dhdr { struct ahdr a; struct bhdr b; struct chdr c;}")
    assert dhdr.fields["a"].fields


def test_parse_type():
    int_ptr = angr.types.parse_type("int *")
    assert isinstance(int_ptr, SimTypePointer)
    assert isinstance(int_ptr.pts_to, SimTypeInt)

    char_ptr = angr.types.parse_type("char *c")
    assert isinstance(char_ptr, SimTypePointer)
    assert isinstance(char_ptr.pts_to, SimTypeChar)

    struct_parse_type = angr.types.parse_type("struct parse_type { char c; float f; }")
    assert isinstance(struct_parse_type, SimStruct)
    assert struct_parse_type.name == "parse_type"
    assert len(struct_parse_type.fields) == 2
    assert isinstance(struct_parse_type.fields["c"], SimTypeChar)
    assert isinstance(struct_parse_type.fields["f"], SimTypeFloat)

    union_dcba = angr.types.parse_type("union dcba { double d; long long int lli; }")
    assert isinstance(union_dcba, SimUnion)
    assert union_dcba.name == "dcba"
    assert len(union_dcba.members) == 2
    assert isinstance(union_dcba.members["d"], SimTypeDouble)
    assert isinstance(union_dcba.members["lli"], SimTypeLongLong)

    struct_llist = angr.types.parse_type("struct llist { int data; struct llist * next; }")
    assert isinstance(struct_llist, SimStruct)
    assert struct_llist.name == "llist"
    assert len(struct_llist.fields) == 2
    assert isinstance(struct_llist.fields["data"], SimTypeInt)
    assert isinstance(struct_llist.fields["next"], SimTypePointer)
    assert isinstance(struct_llist.fields["next"].pts_to, SimStruct)
    assert struct_llist.fields["next"].pts_to.name == "llist"

    func_ptr = angr.types.parse_type("double (*) (int, float)")
    assert isinstance(func_ptr, SimTypePointer)
    assert isinstance(func_ptr.pts_to, SimTypeFunction)
    assert isinstance(func_ptr.pts_to.returnty, SimTypeDouble)
    assert len(func_ptr.pts_to.args) == 2
    assert isinstance(func_ptr.pts_to.args[0], SimTypeInt)
    assert isinstance(func_ptr.pts_to.args[1], SimTypeFloat)


def test_parse_type_no_basic_types():
    time_t = angr.types.parse_type("time_t")
    assert isinstance(time_t, SimTypeLong)

    byte = angr.types.parse_type("byte")
    assert isinstance(byte, SimTypeNum)
    assert byte.size == 8
    assert not byte.signed


def test_self_referential_struct_or_union():
    struct_llist = angr.types.parse_type("struct llist { int data; struct llist *next; }")
    next_struct_llist = struct_llist.fields["next"].pts_to
    assert len(next_struct_llist.fields) == 2
    assert isinstance(next_struct_llist.fields["data"], SimTypeInt)
    assert isinstance(next_struct_llist.fields["next"], SimTypePointer)

    union_heap = angr.types.parse_type("union heap { int data; union heap *forward; }")
    forward_union_heap = union_heap.members["forward"].pts_to
    assert len(forward_union_heap.members) == 2
    assert isinstance(forward_union_heap.members["data"], SimTypeInt)
    assert isinstance(forward_union_heap.members["forward"], SimTypePointer)


def test_union_struct_referencing_each_other():
    angr.types.register_types(angr.types.parse_type("struct a"))
    angr.types.register_types(angr.types.parse_type("struct b"))
    a = angr.types.parse_type("struct a { struct b *b_ptr; }")
    b = angr.types.parse_type("struct b { struct a *a_ptr; }")

    assert len(a.fields) == 1
    assert isinstance(a.fields["b_ptr"], SimTypePointer)
    assert isinstance(a.fields["b_ptr"].pts_to, SimStruct)
    assert a.fields["b_ptr"].pts_to.name == "b"

    assert len(b.fields) == 1
    assert isinstance(b.fields["a_ptr"], SimTypePointer)
    assert isinstance(b.fields["a_ptr"].pts_to, SimStruct)
    assert b.fields["a_ptr"].pts_to.name == "a"

    angr.types.register_types(angr.types.parse_type("union a"))
    angr.types.register_types(angr.types.parse_type("union b"))
    a = angr.types.parse_type("union a { union b *b_ptr; }")
    b = angr.types.parse_type("union b { union a *a_ptr; }")

    assert len(a.members) == 1
    assert isinstance(a.members["b_ptr"], SimTypePointer)
    assert isinstance(a.members["b_ptr"].pts_to, SimUnion)
    assert a.members["b_ptr"].pts_to.name == "b"

    assert len(b.members) == 1
    assert isinstance(b.members["a_ptr"], SimTypePointer)
    assert isinstance(b.members["a_ptr"].pts_to, SimUnion)
    assert b.members["a_ptr"].pts_to.name == "a"


def test_top_type():
    angr.types.register_types({"undefined": angr.types.SimTypeTop()})
    fdef: Dict[str, SimTypeFunction] = angr.types.parse_defns("undefined f(undefined param_1, int param_2);")
    sig = fdef["f"]
    assert sig.args == [angr.types.SimTypeTop(), angr.types.SimTypeInt()]


def test_arg_names():
    angr.types.register_types({"undefined": angr.types.SimTypeTop()})
    fdef: Dict[str, SimTypeFunction] = angr.types.parse_defns("int f(int param_1, int param_2);")
    sig = fdef["f"]
    assert sig.arg_names == ["param_1", "param_2"]

    # Check that arg_names survive a with_arch call
    nsig = sig.with_arch(angr.archinfo.ArchAMD64())
    assert sig.arg_names == nsig.arg_names, "Function type generated with .with_arch() doesn't have identical arg_names"

    # If for some reason only some of the parameters are named,
    # the list can only be partially not None, but has to match the positions
    fdef: Dict[str, SimTypeFunction] = angr.types.parse_defns("int f(int param1, int);")
    sig = fdef["f"]
    assert sig.arg_names == ["param1", None]

    fdef: Dict[str, SimTypeFunction] = angr.types.parse_defns("int f();")
    sig = fdef["f"]
    assert sig.arg_names == ()


def test_varargs():
    fdef = angr.types.parse_defns("int printf(const char *fmt, ...);")
    sig = fdef["printf"]

    assert sig.variadic
    assert "..." in repr(sig)
    assert len(sig.args) == 1
    assert len(sig.arg_names) == 1
    assert "..." not in sig._init_str()


def test_forward_declaration_typedef_struct():
    types, extra_types = angr.types.parse_file("typedef struct _A A; struct _A {int a;int b;};")

    assert extra_types["A"].fields is not None
    assert isinstance(extra_types["A"].fields["a"], SimTypeInt)
    assert isinstance(extra_types["A"].fields["b"], SimTypeInt)

    assert extra_types["struct _A"].fields is not None
    assert isinstance(extra_types["struct _A"].fields["a"], SimTypeInt)
    assert isinstance(extra_types["struct _A"].fields["b"], SimTypeInt)


def test_forward_declaration_typedef_union():
    types, extra_types = angr.types.parse_file("typedef union _A A; union _A {int a;int b;};")

    assert extra_types["A"].members is not None
    assert isinstance(extra_types["A"].members["a"], SimTypeInt)
    assert isinstance(extra_types["A"].members["b"], SimTypeInt)

    assert extra_types["union _A"].members is not None
    assert isinstance(extra_types["union _A"].members["a"], SimTypeInt)
    assert isinstance(extra_types["union _A"].members["b"], SimTypeInt)


def test_bitfield_struct():
    code = """
    struct bitfield_struct {
        uint64_t    qword;
        uint64_t    a    : 36,
                    b     :  8,
                    c  :  7,
                    d      : 12,
                    e      :  1;
        char*       name;
    }"""
    ty = angr.types.parse_type(code)
    ty = ty.with_arch(archinfo.ArchAArch64())
    assert [(t.size, t.offset) for t in list(ty.fields.values())[1:-1]] == [
        (36, 0),
        (8, 4),
        (7, 4),
        (12, 3),
        (1, 7),
    ]


if __name__ == "__main__":
    test_cproto_conversion()
    test_cppproto_conversion()
    test_struct_deduplication()
    test_parse_type()
    test_parse_type_no_basic_types()
    test_self_referential_struct_or_union()
    test_union_struct_referencing_each_other()
    test_top_type()
    test_arg_names()
    test_forward_declaration_typedef_struct()
    test_forward_declaration_typedef_union()
    test_bitfield_struct()



import logging
import os

from angr.calling_conventions import DEFAULT_CC, SimCCUnknown
from angr import SimState, sim_options as o, Project
from archinfo import all_arches, ArchAMD64, ArchSoot

l = logging.getLogger("angr.tests.test_stack_alignment")


def test_alignment():
    for arch in all_arches:
        if arch.name in DEFAULT_CC and DEFAULT_CC[arch.name] is not SimCCUnknown:
            # There is nothing to test for soot about stack alignment
            if isinstance(arch, ArchSoot):
                continue
            l.info("Testing stack alignment for %s", arch.name)
            st = SimState(arch=arch)
            cc = DEFAULT_CC[arch.name](arch=arch)

            st.regs.sp = -1

            # setup callsite with one argument (0x1337), "returning" to 0
            cc.setup_callsite(st, 0, [0x1337], "void foo(int x)")

            # ensure stack alignment is correct
            assert st.solver.is_true((st.regs.sp + cc.STACKARG_SP_DIFF) % cc.STACK_ALIGNMENT == 0), (
                "non-zero stack alignment after setup_callsite for %s" % cc
            )


def test_sys_v_abi_compliance():
    arch = ArchAMD64()
    st = SimState(arch=arch)
    cc = DEFAULT_CC[arch.name](arch=arch)

    st.regs.sp = -1

    # setup callsite with one argument (0x1337), "returning" to 0
    cc.setup_callsite(st, 0, [0x1337], "void foo(int x)")

    # (rsp+8) must be aligned to 16 as required by System V ABI.
    # ref: https://raw.githubusercontent.com/wiki/hjl-tools/x86-psABI/x86-64-psABI-1.0.pdf , page 18t
    assert st.solver.is_true((st.regs.rsp + 8) % 16 == 0), "System V ABI calling convention violated!"


def test_initial_allocation():
    # not strictly about alignment but it's about stack initialization so whatever
    p = Project(
        os.path.join(os.path.dirname(__file__), "../../binaries/tests/x86_64/true"),
        auto_load_libs=False,
    )
    s = p.factory.entry_state(add_options={o.STRICT_PAGE_ACCESS})
    s.memory.load(s.regs.sp - 0x10000, size=4)


if __name__ == "__main__":
    test_alignment()
    test_sys_v_abi_compliance()



import angr
import subprocess
import sys

import logging

l = logging.getLogger("angr.tests.test_signed_div")

from unittest import skipUnless

import os

test_location = os.path.dirname(os.path.realpath(__file__))


@skipUnless(sys.platform.startswith("linux"), "linux only")
def test_signed_div():
    test_bin = os.path.join(test_location, "..", "..", "binaries", "tests", "x86_64", "test_signed_div")
    b = angr.Project(test_bin, auto_load_libs=False)

    pg = b.factory.simulation_manager()
    pg.explore()
    out_angr = pg.deadended[0].posix.dumps(1)
    proc = subprocess.Popen(test_bin, stdout=subprocess.PIPE)
    stdout_real, _ = proc.communicate()

    assert out_angr == stdout_real


if __name__ == "__main__":
    test_signed_div()



import angr

import logging

l = logging.getLogger("angr_tests")

import os

test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


def test_memmove():
    # auto_load_libs can't be disabled as the testcase fails
    proj = angr.Project(
        os.path.join(test_location, "x86_64", "memmove"),
        load_options={"auto_load_libs": True},
        exclude_sim_procedures_list=["memmove"],
    )
    explorer = proj.factory.simulation_manager().explore(find=[0x4005D7])
    s = explorer.found[0]
    result = s.solver.eval(s.memory.load(s.registers.load(16, 8), 13), cast_to=bytes)
    assert result == b"very useful.\x00"


def test_memcpy():
    # auto_load_libs can't be disabled as the testcase fails
    proj = angr.Project(
        os.path.join(test_location, "x86_64", "memcpy"),
        load_options={"auto_load_libs": True},
        exclude_sim_procedures_list=["memcpy"],
    )
    explorer = proj.factory.simulation_manager().explore(find=[0x40065A])
    s = explorer.found[0]
    result = s.solver.eval(s.memory.load(s.registers.load(16, 8), 19), cast_to=bytes)
    assert result == b"let's test memcpy!\x00"


def test_memset():
    # auto_load_libs can't be disabled as the testcase fails
    proj = angr.Project(
        os.path.join(test_location, "x86_64", "memset"),
        load_options={"auto_load_libs": True},
        exclude_sim_procedures_list=["memset"],
    )
    explorer = proj.factory.simulation_manager().explore(find=[0x400608])
    s = explorer.found[0]
    result = s.solver.eval(s.memory.load(s.registers.load(16, 8), 50), cast_to=bytes)
    assert result == b"BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB\x00"


if __name__ == "__main__":
    test_memmove()
    test_memcpy()
    test_memset()



import os

import networkx

import angr
import angr.analyses.decompiler
from angr.analyses.decompiler.structuring import DreamStructurer

test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


class DummyNode:
    def __init__(self, n):
        self.n = n

    def __hash__(self):
        return hash(self.n)

    def __eq__(self, other):
        return isinstance(other, DummyNode) and self.n == other.n or isinstance(other, int) and self.n == other

    @property
    def addr(self):
        return self.n

    def __repr__(self):
        return "<Node %d>" % self.n


def d(n):
    return DummyNode(n)


def D(*edge):
    return DummyNode(edge[0]), DummyNode(edge[1])


def test_region_identifier_0():
    g = networkx.DiGraph()

    #
    #       1
    #       |
    #       2
    #      / \
    #     3  4
    #     \  /
    #      5
    #      |
    #      6

    g.add_edges_from(
        [
            D(1, 2),
            D(2, 3),
            D(2, 4),
            D(3, 5),
            D(4, 5),
            D(5, 6),
        ]
    )

    ri = angr.analyses.decompiler.RegionIdentifier(None, graph=g)
    region = ri.region
    assert len(region.graph.nodes()) == 2


def test_region_identifier_1():
    g = networkx.DiGraph()

    #
    #        1
    #        |
    #        2
    #        | \
    #        | 3
    #        | /
    #        4
    #        |
    #        5
    #        | \
    #        | 6
    #        | /
    #        7
    #        |
    #        8

    g.add_edges_from(
        [
            D(1, 2),
            D(2, 3),
            D(3, 4),
            D(2, 4),
            D(4, 5),
            D(5, 6),
            D(6, 7),
            D(5, 7),
            D(7, 8),
        ]
    )

    ri = angr.analyses.decompiler.RegionIdentifier(None, graph=g)
    region = ri.region
    assert len(region.graph.nodes()) == 2


def test_smoketest():
    p = angr.Project(os.path.join(test_location, "x86_64", "all"), auto_load_libs=False, load_debug_info=True)
    cfg = p.analyses.CFG(data_references=True, normalize=True)

    main_func = cfg.kb.functions["main"]

    # convert function blocks to AIL blocks
    clinic = p.analyses.Clinic(main_func)

    # recover regions
    ri = p.analyses.RegionIdentifier(main_func, graph=clinic.graph)

    # structure it
    st = p.analyses[DreamStructurer].prep()(ri.region)

    # simplify it
    _ = p.analyses.RegionSimplifier(main_func, st.result)


def test_smoketest_cm3_firmware():
    p = angr.Project(
        os.path.join(test_location, "armel", "i2c_master_read-nucleol152re.elf"),
        auto_load_libs=False,
        load_debug_info=True,
    )
    cfg = p.analyses.CFG(normalize=True, force_complete_scan=False)

    main_func = cfg.kb.functions["main"]

    # convert function blocks to AIL blocks
    clinic = p.analyses.Clinic(main_func)

    # recover regions
    ri = p.analyses.RegionIdentifier(main_func, graph=clinic.graph)

    # structure it
    p.analyses[DreamStructurer].prep()(ri.region)


def test_simple():
    p = angr.Project(os.path.join(test_location, "x86_64", "all"), auto_load_libs=False, load_debug_info=True)
    cfg = p.analyses.CFG(data_references=True, normalize=True)

    main_func = cfg.kb.functions["main"]

    # convert function blocks to AIL blocks
    clinic = p.analyses.Clinic(main_func)

    # recover regions
    ri = p.analyses.RegionIdentifier(main_func, graph=clinic.graph)

    # structure it
    rs = p.analyses.RecursiveStructurer(ri.region)

    # simplify it
    s = p.analyses.RegionSimplifier(main_func, rs.result)

    codegen = p.analyses.StructuredCodeGenerator(main_func, s.result, cfg=cfg)
    print(codegen.text)


def test_simple_loop():
    p = angr.Project(
        os.path.join(test_location, "x86_64", "cfg_loop_unrolling"), auto_load_libs=False, load_debug_info=True
    )
    cfg = p.analyses.CFG(data_references=True, normalize=True)

    test_func = cfg.kb.functions["test_func"]

    # convert function blocks to AIL blocks
    clinic = p.analyses.Clinic(test_func)

    # recover regions
    ri = p.analyses.RegionIdentifier(test_func, graph=clinic.graph)

    # structure it
    rs = p.analyses.RecursiveStructurer(ri.region)

    # simplify it
    s = p.analyses.RegionSimplifier(test_func, rs.result)

    codegen = p.analyses.StructuredCodeGenerator(test_func, s.result, cfg=cfg)
    print(codegen.text)

    assert len(codegen.map_pos_to_node._posmap) > 1
    assert len(codegen.map_ast_to_pos) > 1


def test_recursive_structuring():
    p = angr.Project(
        os.path.join(test_location, "x86_64", "cfg_loop_unrolling"), auto_load_libs=False, load_debug_info=True
    )
    cfg = p.analyses.CFG(data_references=True, normalize=True)

    test_func = cfg.kb.functions["test_func"]

    # convert function blocks to AIL blocks
    clinic = p.analyses.Clinic(test_func)

    # recover regions
    ri = p.analyses.RegionIdentifier(test_func, graph=clinic.graph)

    # structure it
    rs = p.analyses.RecursiveStructurer(ri.region)

    # simplify it
    s = p.analyses.RegionSimplifier(test_func, rs.result)

    codegen = p.analyses.StructuredCodeGenerator(test_func, s.result, cfg=cfg)
    print(codegen.text)


def test_while_true_break():
    p = angr.Project(
        os.path.join(test_location, "x86_64", "test_decompiler_loops_O0"), auto_load_libs=False, load_debug_info=True
    )
    cfg = p.analyses.CFG(data_references=True, normalize=True)

    test_func = cfg.kb.functions["_while_true_break"]

    # convert function blocks to AIL blocks
    clinic = p.analyses.Clinic(test_func)

    # recover regions
    ri = p.analyses.RegionIdentifier(test_func, graph=clinic.graph)

    # structure it
    rs = p.analyses.RecursiveStructurer(ri.region)

    # simplify it
    s = p.analyses.RegionSimplifier(test_func, rs.result)

    codegen = p.analyses.StructuredCodeGenerator(test_func, s.result, cfg=cfg)

    print(codegen.text)


def test_while():
    p = angr.Project(
        os.path.join(test_location, "x86_64", "test_decompiler_loops_O0"), auto_load_libs=False, load_debug_info=True
    )
    cfg = p.analyses.CFG(data_references=True, normalize=True)

    test_func = cfg.kb.functions["_while"]

    # convert function blocks to AIL blocks
    clinic = p.analyses.Clinic(test_func)

    # recover regions
    ri = p.analyses.RegionIdentifier(test_func, graph=clinic.graph)

    # structure it
    rs = p.analyses.RecursiveStructurer(ri.region)

    # simplify it
    s = p.analyses.RegionSimplifier(test_func, rs.result)

    codegen = p.analyses.StructuredCodeGenerator(test_func, s.result, cfg=cfg)

    print(codegen.text)


if __name__ == "__main__":
    test_smoketest()
    test_simple()
    test_simple_loop()
    test_recursive_structuring()
    test_while_true_break()
    test_while()
    test_smoketest_cm3_firmware()



import logging

from angr import SimState

l = logging.getLogger("angr.tests.syscalls.mmap")


def test_mmap_base_copy():
    state = SimState(arch="AMD64", mode="symbolic")

    mmap_base = 0x12345678

    state.heap.mmap_base = mmap_base

    # Sanity check
    assert state.heap.mmap_base == mmap_base

    state_copy = state.copy()

    assert state_copy.heap.mmap_base == mmap_base


if __name__ == "__main__":
    g = globals().copy()
    for func_name, func in g.items():
        if func_name.startswith("test_") and hasattr(func, "__call__"):
            func()



from typing import Optional, Tuple, Union, List, DefaultDict, TYPE_CHECKING
from collections import defaultdict
import logging

import networkx

from ailment import Block
from ailment.statement import ConditionalJump
from ailment.expression import Expression, BinaryOp, Const, Load

from ...cfg.cfg_utils import CFGUtils
from ..utils import first_nonlabel_statement, remove_last_statement
from ..structuring.structurer_nodes import IncompleteSwitchCaseHeadStatement, SequenceNode, MultiNode
from ..ailblock_walker import AILBlockWalkerBase
from .optimization_pass import OptimizationPass, OptimizationPassStage, MultipleBlocksException

if TYPE_CHECKING:
    from ailment.expression import UnaryOp, Convert

_l = logging.getLogger(name=__name__)


class Case:
    """
    Describes a case in a switch-case construct.
    """

    __slots__ = (
        "original_node",
        "node_type",
        "variable_hash",
        "expr",
        "value",
        "target",
        "next_addr",
    )

    def __init__(
        self, original_node, node_type: Optional[str], variable_hash, expr, value: Union[int, str], target, next_addr
    ):
        self.original_node = original_node
        self.node_type = node_type
        self.variable_hash = variable_hash
        self.expr = expr
        self.value = value
        self.target = target
        self.next_addr = next_addr

    def __repr__(self):
        if self.value == "default":
            return f"Case default@{self.target:#x}"
        return f"Case {repr(self.original_node)}@{self.target:#x}: {self.expr} == {self.value}"

    def __eq__(self, other):
        if not isinstance(other, Case):
            return False
        return (
            self.original_node == other.original_node
            and self.node_type == other.node_type
            and self.variable_hash == other.variable_hash
            and self.value == other.value
            and self.target == other.target
            and self.next_addr == other.next_addr
        )

    def __hash__(self):
        return hash(
            (Case, self.original_node, self.node_type, self.variable_hash, self.value, self.target, self.next_addr)
        )


class StableVarExprHasher(AILBlockWalkerBase):
    """
    Obtain a stable hash of an AIL expression with respect to all variables and all operations applied on variables.
    """

    def __init__(self, expr: Expression):
        super().__init__()
        self.expr = expr
        self._hash_lst = []

        self.walk_expression(expr)
        self.hash = hash(tuple(self._hash_lst))

    def _handle_expr(self, expr_idx: int, expr: Expression, stmt_idx: int, stmt, block: Optional[Block]):
        if hasattr(expr, "variable"):
            self._hash_lst.append(expr.variable)
        else:
            super()._handle_expr(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt, block: Optional[Block]):
        self._hash_lst.append("Load")
        super()._handle_expr(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_BinaryOp(self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt, block: Optional[Block]):
        self._hash_lst.append(expr.op)
        super()._handle_BinaryOp(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_UnaryOp(self, expr_idx: int, expr: "UnaryOp", stmt_idx: int, stmt, block: Optional[Block]):
        self._hash_lst.append(expr.op)
        super()._handle_UnaryOp(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Const(self, expr_idx: int, expr: Const, stmt_idx: int, stmt, block: Optional[Block]):
        self._hash_lst.append((expr.value, expr.bits))

    def _handle_Convert(self, expr_idx: int, expr: "Convert", stmt_idx: int, stmt, block: Optional[Block]):
        self._hash_lst.append(expr.to_bits)
        super()._handle_Convert(expr_idx, expr, stmt_idx, stmt, block)


class LoweredSwitchSimplifier(OptimizationPass):
    """
    Recognize and simplify lowered switch-case constructs.
    """

    ARCHES = [
        "AMD64",
    ]
    PLATFORMS = ["linux", "windows"]
    STAGE = OptimizationPassStage.BEFORE_REGION_IDENTIFICATION
    NAME = "Convert lowered switch-cases (if-else) to switch-cases"
    DESCRIPTION = (
        "Convert lowered switch-cases (if-else) to switch-cases. Only works when the Phoenix structuring "
        "algorithm is in use."
    )
    STRUCTURING = ["phoenix"]

    def __init__(self, func, blocks_by_addr=None, blocks_by_addr_and_idx=None, graph=None, **kwargs):
        super().__init__(
            func, blocks_by_addr=blocks_by_addr, blocks_by_addr_and_idx=blocks_by_addr_and_idx, graph=graph, **kwargs
        )
        self.analyze()

    def _check(self):
        # TODO: More filtering
        return True, None

    def _analyze(self, cache=None):
        variablehash_to_cases = self._find_cascading_switch_variable_comparisons()

        if not variablehash_to_cases:
            return

        graph_copy = networkx.DiGraph(self._graph)
        self.out_graph = graph_copy
        node_to_heads = defaultdict(set)

        for _, caselists in variablehash_to_cases.items():
            for cases in caselists:
                original_nodes = [case.original_node for case in cases if case.value != "default"]
                original_head: Block = original_nodes[0]
                original_nodes = original_nodes[1:]

                case_addrs = {(case.original_node, case.value, case.target, case.next_addr) for case in cases}

                expr = cases[0].expr

                # create a fake switch-case head node
                switch_stmt = IncompleteSwitchCaseHeadStatement(
                    original_head.statements[-1].idx, expr, case_addrs, ins_addr=original_head.statements[-1].ins_addr
                )
                new_head = original_head.copy()
                # replace the last instruction of the head node with switch_node
                new_head.statements[-1] = switch_stmt
                # update the block
                self._update_block(original_head, new_head)

                # add edges between the head and case nodes
                for onode in original_nodes:
                    successors = list(graph_copy.successors(onode))
                    for succ in successors:
                        if succ not in original_nodes:
                            graph_copy.add_edge(new_head, succ)
                            node_to_heads[succ].add(new_head)
                    graph_copy.remove_node(onode)

        # find shared case nodes and make copies of them
        # note that this only solves cases where *one* node is shared between switch-cases. a more general solution
        # requires jump threading reverter.
        for succ_node, heads in node_to_heads.items():
            if len(heads) > 1:
                # each head gets a copy of the node!
                node_successors = list(graph_copy.successors(succ_node))
                next_id = 0 if succ_node.idx is None else succ_node.idx + 1
                graph_copy.remove_node(succ_node)
                for head in heads:
                    node_copy = succ_node.copy()
                    node_copy.idx = next_id
                    next_id += 1
                    graph_copy.add_edge(head, node_copy)
                    for succ in node_successors:
                        if succ is succ_node:
                            graph_copy.add_edge(node_copy, node_copy)
                        else:
                            graph_copy.add_edge(node_copy, succ)

    def _find_cascading_switch_variable_comparisons(self):
        sorted_nodes = CFGUtils.quasi_topological_sort_nodes(self._graph)
        variable_comparisons = {}
        for node in sorted_nodes:
            r = self._find_switch_variable_comparison_type_a(node)
            if r is not None:
                variable_comparisons[node] = ("a",) + r
                continue
            r = self._find_switch_variable_comparison_type_b(node)
            if r is not None:
                variable_comparisons[node] = ("b",) + r
                continue

        varhash_to_caselists: DefaultDict[int, List[List[Case]]] = defaultdict(list)

        for head in variable_comparisons:
            cases = []
            last_comp = None
            comp = head
            while True:
                comp_type, variable_hash, expr, value, target, next_addr = variable_comparisons[comp]
                if cases:
                    last_varhash = cases[-1].variable_hash
                else:
                    last_varhash = None
                if last_varhash is None or last_varhash == variable_hash:
                    if target == comp.addr:
                        # invalid
                        break
                    cases.append(Case(comp, comp_type, variable_hash, expr, value, target, next_addr))
                else:
                    # new variable!
                    if last_comp is not None:
                        cases.append(Case(last_comp, None, last_varhash, None, "default", comp.addr, None))
                    break

                if comp is not head:
                    # non-head node has at most one predecessor
                    if self._graph.in_degree[comp] > 1:
                        break

                successors = [succ for succ in self._graph.successors(comp) if succ is not comp]
                succ_addrs = {succ.addr for succ in successors}
                if target in succ_addrs:
                    next_comp_addr = next(iter(succ_addr for succ_addr in succ_addrs if succ_addr != target), None)
                    if next_comp_addr is None:
                        break
                    try:
                        next_comp = self._get_block(next_comp_addr)
                    except MultipleBlocksException:
                        # multiple blocks :/ it's possible that other optimization passes have duplicated the default
                        # node. check it.
                        next_comp_many = list(self._get_blocks(next_comp_addr))
                        if next_comp_many[0] not in variable_comparisons:
                            cases.append(Case(comp, None, variable_hash, expr, "default", next_comp_addr, None))
                        # otherwise we don't support it
                        break
                    assert next_comp is not None
                    if next_comp in variable_comparisons:
                        last_comp = comp
                        comp = next_comp
                        continue
                    cases.append(Case(comp, None, variable_hash, expr, "default", next_comp_addr, None))
                break

            if cases:
                v = cases[-1].variable_hash
                for idx, existing_cases in list(enumerate(varhash_to_caselists[v])):
                    if self.cases_issubset(existing_cases, cases):
                        varhash_to_caselists[v][idx] = cases
                        break
                    if self.cases_issubset(cases, existing_cases):
                        break
                else:
                    varhash_to_caselists[v].append(cases)

        for v, caselists in list(varhash_to_caselists.items()):
            for idx, cases in list(enumerate(caselists)):
                # filter: there should be at least two non-default cases
                if len([case for case in cases if case.value != "default"]) < 2:
                    caselists[idx] = None
                    continue

                # filter: no type-a node after the first case node
                if any(case for case in cases[1:] if case.value != "default" and case.node_type == "a"):
                    caselists[idx] = None
                    continue

                # filter: each case is only reachable from a case node
                all_case_nodes = {case.original_node for case in cases}
                skipped = False
                for case in cases:
                    target_nodes = [
                        succ for succ in self._graph.successors(case.original_node) if succ.addr == case.target
                    ]
                    if len(target_nodes) != 1:
                        caselists[idx] = None
                        skipped = True
                        break
                    target_node = target_nodes[0]
                    nonself_preds = {pred for pred in self._graph.predecessors(target_node) if pred.addr == case.target}
                    if not nonself_preds.issubset(all_case_nodes):
                        caselists[idx] = None
                        skipped = True
                        break

                if skipped:
                    continue

            varhash_to_caselists[v] = [cl for cl in caselists if cl is not None]

        return varhash_to_caselists

    @staticmethod
    def _find_switch_variable_comparison_type_a(
        node,
    ) -> Optional[Tuple[int, Expression, int, int, int]]:
        # the type a is the last statement is a var == constant comparison, but
        # there is more than one non-label statement in the block

        if isinstance(node, Block) and node.statements:
            stmt = node.statements[-1]
            if stmt is not None and stmt is not first_nonlabel_statement(node):
                if (
                    isinstance(stmt, ConditionalJump)
                    and isinstance(stmt.true_target, Const)
                    and isinstance(stmt.false_target, Const)
                ):
                    cond = stmt.condition
                    if isinstance(cond, BinaryOp):
                        if isinstance(cond.operands[1], Const):
                            variable_hash = StableVarExprHasher(cond.operands[0]).hash
                            value = cond.operands[1].value
                            if cond.op == "CmpEQ":
                                target = stmt.true_target.value
                                next_node_addr = stmt.false_target.value
                            elif cond.op == "CmpNE":
                                target = stmt.false_target.value
                                next_node_addr = stmt.true_target.value
                            else:
                                return None
                            return variable_hash, cond.operands[0], value, target, next_node_addr

        return None

    @staticmethod
    def _find_switch_variable_comparison_type_b(
        node,
    ) -> Optional[Tuple[int, Expression, int, int, int]]:
        # the type b is the last statement is a var == constant comparison, and
        # there is only one non-label statement

        if isinstance(node, Block):
            stmt = first_nonlabel_statement(node)
            if stmt is not None and stmt is node.statements[-1]:
                if (
                    isinstance(stmt, ConditionalJump)
                    and isinstance(stmt.true_target, Const)
                    and isinstance(stmt.false_target, Const)
                ):
                    cond = stmt.condition
                    if isinstance(cond, BinaryOp):
                        if isinstance(cond.operands[1], Const):
                            variable_hash = StableVarExprHasher(cond.operands[0]).hash
                            value = cond.operands[1].value
                            if cond.op == "CmpEQ":
                                target = stmt.true_target.value
                                next_node_addr = stmt.false_target.value
                            elif cond.op == "CmpNE":
                                target = stmt.false_target.value
                                next_node_addr = stmt.true_target.value
                            else:
                                return None
                            return variable_hash, cond.operands[0], value, target, next_node_addr

        return None

    @staticmethod
    def restore_graph(
        node, last_stmt: IncompleteSwitchCaseHeadStatement, graph: networkx.DiGraph, full_graph: networkx.DiGraph
    ):
        last_node = node
        ca_default = [
            (onode, value, target, a) for onode, value, target, a in last_stmt.case_addrs if value == "default"
        ]
        ca_others = [
            (onode, value, target, a) for onode, value, target, a in last_stmt.case_addrs if value != "default"
        ]

        # non-default nodes
        ca_others = {ca[0].addr: ca for ca in ca_others}
        # extract the AIL block from last_node
        last_block = last_node
        if isinstance(last_block, SequenceNode):
            last_block = last_block.nodes[-1]
        if isinstance(last_block, MultiNode):
            last_block = last_block.nodes[-1]
        assert isinstance(last_block, Block)
        next_node_addr = last_block.addr

        while next_node_addr is not None and next_node_addr in ca_others:
            onode, value, target, next_node_addr = ca_others[next_node_addr]
            onode: Block

            if first_nonlabel_statement(onode) is not onode.statements[-1]:
                onode = onode.copy(statements=[onode.statements[-1]])

            graph.add_edge(last_node, onode)
            full_graph.add_edge(last_node, onode)

            target_node = next(iter(nn for nn in full_graph if nn.addr == target))
            graph.add_edge(onode, target_node)
            full_graph.add_edge(onode, target_node)

            if graph.has_edge(node, target_node):
                graph.remove_edge(node, target_node)
            if full_graph.has_edge(node, target_node):
                full_graph.remove_edge(node, target_node)

            # update last_node
            last_node = onode

        # default nodes
        if ca_default:
            onode, value, target, _ = ca_default[0]
            default_target = next(iter(nn for nn in full_graph if nn.addr == target))
            graph.add_edge(last_node, default_target)
            full_graph.add_edge(last_node, default_target)

            if graph.has_edge(node, default_target):
                graph.remove_edge(node, default_target)
            if full_graph.has_edge(node, default_target):
                full_graph.remove_edge(node, default_target)

        # all good - remove the last statement in node
        remove_last_statement(node)

    @staticmethod
    def cases_issubset(cases_0: List[Case], cases_1: List[Case]) -> bool:
        """
        Test if cases_0 is a subset of cases_1.
        """

        if len(cases_0) > len(cases_1):
            return False
        for case in cases_0:
            if case not in cases_1:
                return False
        return True



import angr
import os

test_location = os.path.join(os.path.dirname(os.path.realpath(str(__file__))), "..", "..", "binaries", "tests", "")


def test_static_hooker():
    test_file = os.path.join(test_location, "x86_64", "static")
    p = angr.Project(test_file, auto_load_libs=False)
    sh = p.analyses.StaticHooker("libc.so.6")

    assert 4197616 in sh.results
    assert type(sh.results[4197616]) is angr.SIM_PROCEDURES["glibc"]["__libc_start_main"]
    assert type(p.hooked_by(4197616)) is angr.SIM_PROCEDURES["glibc"]["__libc_start_main"]


if __name__ == "__main__":
    test_static_hooker()



import angr

from cle.backends.externs.simdata.io_file import io_file_data_for_arch

######################################
# fputc
######################################


class fputc(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, c, file_ptr):
        fd_offset = io_file_data_for_arch(self.state.arch)["fd"]
        fileno = self.state.mem[file_ptr + fd_offset :].int.resolved
        simfd = self.state.posix.get_fd(fileno)
        if simfd is None:
            return -1

        simfd.write_data(c[7:0])
        return c & 0xFF


putc = fputc
fputc_unlocked = fputc
putc_unlocked = fputc



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("api-ms-win-gaming-tcui-l1-1-2.dll")
prototypes = \
    {
        #
        'ShowGameInviteUIForUser': SimTypeFunction([SimTypeBottom(label="IInspectable"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["returnCode", "context"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["user", "serviceConfigurationId", "sessionTemplateName", "sessionId", "invitationDisplayText", "completionRoutine", "context"]),
        #
        'ShowPlayerPickerUIForUser': SimTypeFunction([SimTypeBottom(label="IInspectable"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="HSTRING"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypeBottom(label="Void"), arg_names=["returnCode", "context", "selectedXuids", "selectedXuidsCount"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["user", "promptDisplayText", "xuids", "xuidsCount", "preSelectedXuids", "preSelectedXuidsCount", "minSelectionCount", "maxSelectionCount", "completionRoutine", "context"]),
        #
        'ShowProfileCardUIForUser': SimTypeFunction([SimTypeBottom(label="IInspectable"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["returnCode", "context"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["user", "targetUserXuid", "completionRoutine", "context"]),
        #
        'ShowChangeFriendRelationshipUIForUser': SimTypeFunction([SimTypeBottom(label="IInspectable"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["returnCode", "context"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["user", "targetUserXuid", "completionRoutine", "context"]),
        #
        'ShowTitleAchievementsUIForUser': SimTypeFunction([SimTypeBottom(label="IInspectable"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["returnCode", "context"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["user", "titleId", "completionRoutine", "context"]),
        #
        'CheckGamingPrivilegeWithUIForUser': SimTypeFunction([SimTypeBottom(label="IInspectable"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["returnCode", "context"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["user", "privilegeId", "scope", "policy", "friendlyMessage", "completionRoutine", "context"]),
        #
        'CheckGamingPrivilegeSilentlyForUser': SimTypeFunction([SimTypeBottom(label="IInspectable"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["user", "privilegeId", "scope", "policy", "hasPrivilege"]),
    }

lib.set_prototypes(prototypes)



# pylint: disable=missing-class-docstring,disable=no-self-use
import os
import unittest

import angr

test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


class TestFindObjectsStatic(unittest.TestCase):
    def test_object_identification_x86_64(self):
        p = angr.Project(os.path.join(test_location, "x86_64", "cpp_classes"), auto_load_libs=False)
        object_identifier_analysis = p.analyses.StaticObjectFinder()
        possible_objects_dict = object_identifier_analysis.possible_objects
        possible_constructors = object_identifier_analysis.possible_constructors
        class_labels = []

        for possible_object in possible_objects_dict.values():
            class_labels.append(possible_object.class_name)

        assert "C" in class_labels
        assert len(possible_objects_dict) == 2
        assert len(possible_constructors) == 1
        assert 0x401512 in possible_constructors
        assert len(possible_constructors[0x401512]) == 2


if __name__ == "__main__":
    unittest.main()



import os
import sys

import angr

test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


def test_various_loops():
    p = angr.Project(os.path.join(test_location, "x86_64", "various_loops"), auto_load_libs=False)

    cfg = p.analyses.CFGFast(normalize=True)

    state = p.factory.entry_state()
    state.register_plugin("loop_data", angr.state_plugins.SimStateLoopData())

    dummy = p.loader.main_object.get_symbol("dummy")
    bvs = state.solver.BVS(dummy.name, 8 * dummy.size)
    state.memory.store(dummy.rebased_addr, bvs, endness="Iend_LE")

    simgr = p.factory.simulation_manager(state)

    simgr.use_technique(angr.exploration_techniques.LoopSeer(cfg=cfg, functions=None, bound=None))

    simgr.run()

    assert len(simgr.deadended) == 10
    assert len(simgr.deadended[0].loop_data.back_edge_trip_counts) == 14

    for i, d in enumerate(simgr.deadended):
        f = p.kb.functions.function(name="symbolic_loop")
        l = p.analyses.LoopFinder(functions=[f]).loops[0]
        assert d.loop_data.back_edge_trip_counts[l.entry.addr][0] == i

        f = p.kb.functions.function(name="for_loop")
        l = p.analyses.LoopFinder(functions=[f]).loops[0]
        assert d.loop_data.back_edge_trip_counts[l.entry.addr][0] == 9

        f = p.kb.functions.function(name="while_loop")
        l = p.analyses.LoopFinder(functions=[f]).loops[0]
        assert d.loop_data.back_edge_trip_counts[l.entry.addr][0] == 9

        f = p.kb.functions.function(name="do_while_loop")
        l = p.analyses.LoopFinder(functions=[f]).loops[0]
        assert d.loop_data.header_trip_counts[l.entry.addr][0] == 9

        f = p.kb.functions.function(name="nullify")
        l = p.analyses.LoopFinder(functions=[f]).loops[0]
        assert len(d.loop_data.back_edge_trip_counts[l.entry.addr]) == 8
        assert d.loop_data.back_edge_trip_counts[l.entry.addr][0] == 9

        f = p.kb.functions.function(name="nested_for_loop")
        ol = p.analyses.LoopFinder(functions=[f]).loops[0]
        il = ol.subloops[0]
        assert d.loop_data.back_edge_trip_counts[ol.entry.addr][0] == 3
        assert len(d.loop_data.back_edge_trip_counts[il.entry.addr]) == 3
        assert all(s == 3 for s in d.loop_data.back_edge_trip_counts[il.entry.addr])

        f = p.kb.functions.function(name="nested_while_loop")
        ol = p.analyses.LoopFinder(functions=[f]).loops[0]
        il = ol.subloops[0]
        assert d.loop_data.back_edge_trip_counts[ol.entry.addr][0] == 3
        assert len(d.loop_data.back_edge_trip_counts[il.entry.addr]) == 3
        assert all(s == 3 for s in d.loop_data.back_edge_trip_counts[il.entry.addr])

        f = p.kb.functions.function(name="nested_do_while_loop")
        ol = p.analyses.LoopFinder(functions=[f]).loops[0]
        il = ol.subloops[0]
        assert d.loop_data.header_trip_counts[ol.entry.addr][0] == 3
        assert len(d.loop_data.header_trip_counts[il.entry.addr]) == 3
        assert all(s == 3 for s in d.loop_data.header_trip_counts[il.entry.addr])

        f = p.kb.functions.function(name="break_for_loop")
        l = p.analyses.LoopFinder(functions=[f]).loops[0]
        assert d.loop_data.back_edge_trip_counts[l.entry.addr][0] == 9

        f = p.kb.functions.function(name="break_do_while_loop")
        l = p.analyses.LoopFinder(functions=[f]).loops[0]
        assert d.loop_data.header_trip_counts[l.entry.addr][0] == 9


def test_loops_with_invalid_parameter():
    p = angr.Project(os.path.join(test_location, "x86_64", "test_loops"), auto_load_libs=False)

    state = p.factory.entry_state()
    state.register_plugin("loop_data", angr.state_plugins.SimStateLoopData())
    simgr = p.factory.simulation_manager(state)

    simgr.use_technique(angr.exploration_techniques.LoopSeer(functions=["main", 0x1234], bound=None))

    simgr.run()

    assert len(simgr.deadended[0].loop_data.back_edge_trip_counts) == 3
    assert simgr.deadended[0].loop_data.back_edge_trip_counts[0x400665][0] == 10
    assert len(simgr.deadended[0].loop_data.back_edge_trip_counts[0x400665]) == 10
    assert simgr.deadended[0].loop_data.back_edge_trip_counts[0x400675][0] == 10
    assert simgr.deadended[0].loop_data.back_edge_trip_counts[0x4006B2][0] == 100


def test_arrays():
    p = angr.Project(os.path.join(test_location, "x86_64", "test_arrays"), auto_load_libs=False)

    cfg = p.analyses.CFGFast(normalize=True)

    state = p.factory.entry_state()
    state.register_plugin("loop_data", angr.state_plugins.SimStateLoopData())
    simgr = p.factory.simulation_manager(state)

    simgr.use_technique(angr.exploration_techniques.LoopSeer(cfg=cfg, functions="main", bound=None))

    simgr.run()

    assert len(simgr.deadended[0].loop_data.back_edge_trip_counts) == 2
    assert simgr.deadended[0].loop_data.back_edge_trip_counts[0x400636][0] == 26
    assert simgr.deadended[0].loop_data.back_edge_trip_counts[0x4005FD][0] == 26


def test_loop_limiter():
    p = angr.Project(os.path.join(test_location, "x86_64", "test_arrays"), auto_load_libs=False)

    cfg = p.analyses.CFGFast(normalize=True)

    state = p.factory.entry_state()
    state.register_plugin("loop_data", angr.state_plugins.SimStateLoopData())
    simgr = p.factory.simulation_manager(state)

    simgr.use_technique(angr.exploration_techniques.LoopSeer(cfg=cfg, functions="main", bound=5))

    simgr.run()

    assert "spinning" in simgr.stashes
    assert simgr.spinning[0].loop_data.back_edge_trip_counts[0x4005FD][0] == 6


def test_loop_limiter_constant_loop():
    p = angr.Project(os.path.join(test_location, "x86_64", "constant_loopseer"), auto_load_libs=False)

    cfg = p.analyses.CFGFast(normalize=True)

    state = p.factory.entry_state()
    simgr = p.factory.simulation_manager(state)

    simgr.use_technique(
        angr.exploration_techniques.LoopSeer(cfg=cfg, functions="main", bound=5, limit_concrete_loops=False)
    )

    simgr.run()
    assert simgr.deadended[0].regs.eax.concrete
    val = simgr.deadended[0].solver.eval_one(simgr.deadended[0].regs.eax)
    assert val == 420


if __name__ == "__main__":
    if len(sys.argv) > 1:
        globals()["test_" + sys.argv[1]]()

    else:
        g = globals().copy()

        for k, v in g.items():
            if k.startswith("test_") and hasattr(v, "__call__"):
                print(k)
                v()



import angr

from cle.backends.externs.simdata.io_file import io_file_data_for_arch

######################################
# fdopen
#
# Reference for implementation:
#   glibc-2.25/libio/iofdopen.c
######################################


def mode_to_flag(mode):
    # TODO improve this: handle mode = strings
    if mode[-1] == b"b":  # lol who uses windows
        mode = mode[:-1]
    all_modes = {
        b"r": angr.storage.file.Flags.O_RDONLY,
        b"r+": angr.storage.file.Flags.O_RDWR,
        b"w": angr.storage.file.Flags.O_WRONLY | angr.storage.file.Flags.O_CREAT,
        b"w+": angr.storage.file.Flags.O_RDWR | angr.storage.file.Flags.O_CREAT,
        b"a": angr.storage.file.Flags.O_WRONLY | angr.storage.file.Flags.O_CREAT | angr.storage.file.Flags.O_APPEND,
        b"a+": angr.storage.file.Flags.O_RDWR | angr.storage.file.Flags.O_CREAT | angr.storage.file.Flags.O_APPEND,
    }
    if mode not in all_modes:
        raise angr.SimProcedureError("unsupported file open mode %s" % mode)

    return all_modes[mode]


class fdopen(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, fd_int, m_addr):
        # pylint:disable=unused-variable
        strlen = angr.SIM_PROCEDURES["libc"]["strlen"]

        m_strlen = self.inline_call(strlen, m_addr)
        m_expr = self.state.memory.load(m_addr, m_strlen.max_null_index, endness="Iend_BE")
        self.state.solver.eval(m_expr, cast_to=bytes)

        # TODO: handle append and other mode subtleties

        fd = self.state.solver.eval(fd_int)
        if fd not in self.state.posix.fd:
            # if file descriptor not found return NULL
            return 0
        else:
            # Allocate a FILE struct in heap
            malloc = angr.SIM_PROCEDURES["libc"]["malloc"]
            io_file_data = io_file_data_for_arch(self.state.arch)
            file_struct_ptr = self.inline_call(malloc, io_file_data["size"]).ret_expr

            # Write the fd
            fd_bvv = self.state.solver.BVV(fd, 4 * 8)  # int
            self.state.memory.store(
                file_struct_ptr + io_file_data["fd"], fd_bvv, endness=self.state.arch.memory_endness
            )

            return file_struct_ptr
